##############################################################################
# Author: Bernard Hernandez
# Filename: wls.py
#
# Description : This file contains a wrapper for the weighted least squares
#               prediction model implemented in statsmodels.
#
###############################################################################
# Import future
from __future__ import division

# Libraries
import sys
import copy
import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm

# Import specific
from statsmodels.sandbox.regression.predstd import wls_prediction_std

# Libraries wrapper.
from pyamr.core.regression.wreg import RegressionWrapper
from pyamr.core.stats.wbase import fargs


# --------------------------------------------------------------------------
#                              helper methods
# -------------------------------------------------------------------------- 
def _cast_float(e):
    """Casts an element to float when feasible.
    """
    try:
        return float(e)
    except:
        return e


class PredictionResult():

    def __init__(self, mean, cilo, ciup, pstd, pilo,
                 piup, time, nobs, endog):
        """The constructor

        Parameters
        ----------

        Returns
        -------
        """
        # Set variables
        self.mean = mean
        self.cilo = cilo
        self.ciup = ciup
        self.pstd = pstd
        self.pilo = pilo
        self.piup = piup
        self.time = time
        self.nobs = nobs
        self.endog = endog

        # Set combined interval
        self.bilo, self.biup = np.copy(cilo), np.copy(ciup)
        self.bilo[self.nobs:] = self.pilo[self.nobs:]
        self.biup[self.nobs:] = self.piup[self.nobs:]


class WLSWrapper(RegressionWrapper):
    """The description...

    """

    # Attributes.
    _name = 'WLS'  # Label to add to the attributes when saving.

    # --------------------------------------------------------------------------
    #                           helper methods
    # --------------------------------------------------------------------------
    def _identifier(self, **kwargs):
        """This methods describes de model."""
        # Returns description
        if self.W is not None:
            if hasattr(self.W, '_identifier'):
                return "%s(%s,%s)" % (self._name,
                                      self.trend,
                                      self.W._identifier(**kwargs))
            else:
                return "%s(%s,%s)" % (self._name,
                                      self.trend,
                                      self.W.__class__.__name__)
        # Return basic
        return "%s(%s)" % (self._name, self.trend)

    def pred_int(self, start=None, end=None, **kwargs):
        """This method computes the prediction intervals

        Parameters
        ----------
        start : int (optional)
          The time t to start the prediction

        end : int (optional)
          The time t to end the prediction

        Returns
        -------
        the standard prediction error and the prediction intervals
        """
        # Create the exogenous variable for prediction
        exog = self._exog(start, end)
        # Compute
        pstd, pilo, piup = wls_prediction_std(self._raw, exog=exog, **kwargs)
        # Retrn
        return np.column_stack((pilo, piup))

    # --------------------------------------------------------------------------
    #                            set variables
    # --------------------------------------------------------------------------
    def _params_from_summary(self):
        """Gets parameters from the summary result of the raw object.

        .. note: There must be a way to get all the elements extracted from
                 the summary using the raw object and the methods implemented
                 within.

                 See: https://www.statsmodels.org/dev/generated/statsmodels.regression.linear_model.RegressionResults.html
        """
        # Format summary
        summary = self._raw.summary().as_csv()
        summary = summary.split("\n", 1)[1]  # Remove first line.
        summary = summary.replace("\n", ",")  # Replace \n by comma.

        # Split in elements.
        elements = summary.split(",")
        elements = [_cast_float(e.strip()) for e in elements]


        #print(self._raw.summary())
        #for tb in self._raw.summary().tables:
        #    b = pd.read_html(tb.as_html(), header=0, index_col=0)[0]
        #    print(b)
        #    print("\n")

        #import sys
        #sys.exit()

        # Create series.
        d = {}

        # Add parameters.
        d['s_dw'] = elements[61]
        d['s_jb_value'] = elements[65]
        d['s_jb_prob'] = elements[69]
        d['s_skew'] = elements[67]
        d['s_kurtosis'] = elements[71]
        d['s_omnibus_value'] = elements[59]
        d['s_omnibus_prob'] = elements[63]

        # Return
        return d

    def evaluate(self, alpha=0.05):
        """This method set all the variables into this class.

        Notes:
          - if instead of having the attribute series it is desired to set
            each element of the series as an attribute, just used the following
            statement: setattr(self, name, value).

          - Note the difference between the resid and wresid since they will
            also have different statistical properties. Furthermore, there is
            a vector with normalized residuals resid_pearson.

        @see: statsmodels.WLS
        @see: pyAMR.core.regression.RegressionResultsWrapper

        Parameters
        ----------
        alpha : the confidence interval

        Returns
        -------
        dictionary : map with all the parameters.
        """
        # Libraries.
        from statsmodels.iolib.summary import _getnames

        # Create series.
        d = {}

        # Add generic metrics.
        d['rsquared'] = self._raw.rsquared
        d['rsquared_adj'] = self._raw.rsquared_adj
        d['fvalue'] = self._raw.fvalue
        d['fprob'] = self._raw.f_pvalue
        d['aic'] = self._raw.aic
        d['bic'] = self._raw.bic
        d['llf'] = self._raw.llf
        d['mse_model'] = self._raw.mse_model
        d['mse_resid'] = self._raw.mse_resid
        d['mse_total'] = self._raw.mse_total

        # Create params information.
        params_data = zip(_getnames(self._raw)[1],
                          self._raw.params,
                          self._raw.bse,
                          self._raw.tvalues,
                          self._raw.pvalues,
                          self._raw.conf_int(alpha))

        # Add coefficients statistics.
        for name, coef, std, tvalue, pvalue, (cil, ciu) in params_data:
            d['%s_%s' % (name, 'coef')] = coef
            d['%s_%s' % (name, 'std')] = std
            d['%s_%s' % (name, 'tvalue')] = tvalue
            d['%s_%s' % (name, 'tprob')] = pvalue
            d['%s_%s' % (name, 'cil')] = cil
            d['%s_%s' % (name, 'ciu')] = ciu

        # Further statistics.
        d.update(self._params_from_summary())
        d.update(self._resid_stats())

        # Return
        return d

    # --------------------------------------------------------------------------
    #                           display methods
    # --------------------------------------------------------------------------
    def as_summary(self, **kwargs):
        """This method creates a summary string.
        """
        # Elements to split by.
        find = "=" * 78
        # Split and fill.
        smry = find.join(self._raw.summary(**kwargs).as_text().split(find)[:-1])
        # Add in new lines.
        smry += "Normal (N): %#25.3f   Prob(N): %#29.3f\n" % (self.m_nm_value,
                                                              self.m_nm_prob)
        smry += find
        # Return
        return smry

    # --------------------------------------------------------------------------
    #                          fit and predict methods
    # --------------------------------------------------------------------------
    def fit(self, endog, exog=None, trend='n',
            weights=None,
            W=None,
            **kwargs):
        """This method computes the WLS.

        @see statsmodels.regression.linear_model.WLS

        Parameters
        ----------
        endog: array-like
          The endogenous variable (aka. time series data)

        exog: array-like
          The exogenous variable (by default is the time t starting in 0)

        trend:  str-like, options = {c, n}
          Wether to add a constant or not.

        weights: array-like (optional)
          The weights for the weighted least square regression. If weights and
          W are both not None, the W instance will be used to transform the
          weights variables.

        W: object-like (optional)
          The instance to transform the weights. It must implement the
          function 'weights'.

        kwargs: dict-like
          The rest of the arguments to pass to the  __init__ and fit methods
          of the class statsmodels.WLS (see xxx)

        Returns
        -------
        object : OLSWrapper object.
        """

        # Create a time-series exogenous variable
        if exog is None:
            exog = np.arange(endog.size)

        # Format the exogenous variable to add constant.
        if trend == 'c':
            exog = sm.add_constant(copy.deepcopy(exog))

        # Compute weights from frequency
        if weights is not None and W is not None:
            weights = W.weights(weights)
            if np.isnan(weights).any():
                warnings.warn("""\n
                     There was an error computing the weights. In order to
                     avoid a fatal error, uniform weights will be set. The 
                     weight transformer used was: {}""".format(W))
                weights = None

        # Set uniform weights
        if weights is None :
            weights = np.ones(endog.size)

        # Save all newly created parameters
        kwargs['exog'] = exog
        kwargs['endog'] = endog
        kwargs['trend'] = trend
        kwargs['weights'] = weights
        kwargs['W'] = W


        # Call parents fit.
        super(WLSWrapper, self).fit(**kwargs)

        # Save results.
        return self

    def get_prediction(self, start=None, end=None, alpha=0.05, **kwargs):
        """This method predicts using the model.

        .. todo::

            Note that wls_predict_std has weights (those used in WLS) as
            an input parameter. However, these are not passed to the
            function. Those weights are available for insample
            predictions but not for forecasting.

        Parameters
        ----------
        start : int (optional)
          The time t to start the prediction

        end : int (optional)
          The time t to end the prediction

        kwargs:
          The arguments to pass to the method get_prediction of the
          class statsmodels.WLS (see xxx)

        Returns
        -------
        time, prediction mean, prediction confidence interval
        """
        # Create exogenous variable.
        if start is not None or end is not None:
            kwargs['exog'] = self._exog(start, end)

        # Compute prediction.
        prediction = self._raw.get_prediction(**kwargs)

        # Add time.
        time = self._time(start=start, end=end)

        # Get plotting values
        mean = prediction.predicted_mean
        cito = prediction.conf_int()
        pito = self.pred_int(start=start, end=end, alpha=alpha)

        # Number of insample prediction
        ninsample = np.sum(time < self.endog.size)

        # Fill insamples
        pito[:ninsample, :] = cito[:ninsample, :]

        # Create result
        return np.column_stack((time, mean, pito)).T

    def line(self, x):
        """This method returns arrays to plot line and confidence intervals.
        """
        if 'const_coef' in self._result:
            return self.x1_coef * x + self.const_coef
        else:
            return self.x1_coef * x


if __name__ == '__main__': # pragma: no cover

    # Import
    import sys
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import statsmodels.robust.norms as norms

    # import weights.
    from pyamr.datasets.load import make_timeseries
    from pyamr.metrics.weights import SigmoidA

    # ----------------------------
    # set basic configuration
    # ----------------------------
    # Matplotlib options
    mpl.rc('legend', fontsize=6)
    mpl.rc('xtick', labelsize=6)
    mpl.rc('ytick', labelsize=6)

    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 150)
    pd.set_option('display.precision', 4)

    # ----------------------------
    # create data
    # ----------------------------
    # Create timeseries data
    x, y, f = make_timeseries()

    # Create method to compute weights from frequencies
    W = SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0)

    # ----------------------------
    # fit the model
    # ----------------------------
    # Note that the function fit will call M.weights(weights) inside and will
    # store the M converter in the instance. Therefore, the code execute is
    # equivalent to <weights=M.weights(f)> with the only difference being that
    # the weight converter is not saved.
    wls = WLSWrapper(estimator=sm.WLS).fit(exog=x[:80],
                                           endog=y[:80],
                                           trend='c',
                                           weights=f[:80],
                                           W=W,
                                           missing='raise')

    # Print series.
    print("\nSeries:")
    print(wls.as_series())

    # Print regression line.
    print("\nRegression line:")
    print(wls.line(np.arange(10)))

    # Print summary.
    print("\nSummary:")
    print(wls.as_summary())

    # -----------------
    # Save & Load
    # -----------------
    # File location
    # fname = '../../examples/saved/wls-sample.pickle'

    # Save
    # wls.save(fname=fname)

    # Load
    # wls = WLSWrapper().load(fname=fname)

    # -------------
    #  Example I
    # -------------
    # This example shows how to make predictions using the wrapper and how
    # to plot the resulting data. In addition, it compares the intervales
    # provided by get_prediction (confidence intervals) and the intervals
    # provided by wls_prediction_std (prediction intervals).
    #
    # To Do: Implement methods to compute CI and PI (see regression).

    # Variables.
    start, end = None, 180

    # Compute predictions (exogenous?).
    preds = wls.get_prediction(start=start, end=end)

    # Create figure
    fig, ax = plt.subplots(1, 1, figsize=(11, 5))

    # Plotting confidence intervals
    # -----------------------------
    # Plot truth values.
    ax.plot(x, y, color='#A6CEE3', alpha=0.5, marker='o',
            markeredgecolor='k', markeredgewidth=0.5,
            markersize=5, linewidth=0.75, label='Observed')

    # Plot forecasted values.
    ax.plot(preds[0, :], preds[1, :], color='#FF0000', alpha=1.00,
            linewidth=2.0, label=wls._identifier(short=True))

    # Plot the confidence intervals.
    ax.fill_between(preds[0, :], preds[2, :],
                    preds[3, :],
                    color='r',
                    alpha=0.1)

    # Legend
    plt.legend()

    # -----------------------------
    # Example II
    # -----------------------------
    # This example performs grid search on a number of possible configurations
    # of the WLSWrapper. In particular, it tests the effect of different
    # objects to compute the weights from the frequencies. It presents both
    # the resulting pandas dataframe and also a figure.

    # Configuration
    # -------------
    # This variable contains the weight functions to test. Note that in
    # the norms module there are other options such as [norms.HuberT(),
    # norms.Hampel(), norms.TrimmedMean(), norms.TukeyBiweight(),
    # norms.AndreWave(), norms.RamsayE()]
    w_func = [
        norms.LeastSquares(),
        SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0),
        SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[10, 90]),
        SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[25, 75]),
        SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[25, 90]),
        SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[40, 50])]

    # The grid search parameters.
    grid_params = [
        # {'exog': [x], 'endog': [y], 'trend': ['c']},
        {'exog': [x], 'endog': [y], 'trend': ['c'], 'weights': [f], 'W': w_func}
    ]

    # Grid search
    # ------------
    # Perform grid search.
    summary = WLSWrapper(estimator=sm.WLS).grid_search(grid_params=grid_params)

    # Show grid results
    print("\nGrid search:")
    print(WLSWrapper().from_list_dataframe(summary).T)

    # Prediction
    # ----------
    # Variables.
    start, end = 10, 150

    # Create figure
    fig, axes = plt.subplots(1, 3, figsize=(10, 5))

    # Plot truth values.
    axes[0].plot(x, y, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=5, linewidth=0.75, label='Observed')

    # Plot frequencies
    axes[0].bar(x, f, color='gray', alpha=0.7, label='Frequency')

    # For each of the models in summary
    for i, model in enumerate(summary):

        # Compute predictions.
        preds = model.get_prediction(start=start, end=end)

        # Plot forecasted values.
        axes[0].plot(preds[0, :], preds[1, :], linewidth=1.0,
                     label=model._identifier(short=True))

        # Plot the confidence intervals.
        axes[0].fill_between(preds[0, :], preds[2, :], preds[3, :], alpha=0.1)

        # Plot weights assigned to each observation
        axes[1].plot(model.weights, marker='o', alpha=0.5,
                     markeredgecolor='k', markeredgewidth=0.5,
                     markersize=4, linewidth=0.00,
                     label=model._identifier(short=True))

        # Plot weights converter (W) functions.
        if model.W is not None:
            axes[2].plot(np.linspace(0, 1, 100),
                         model.W.weights(np.linspace(0, 1, 100)),
                         label=model._identifier(short=True))

    # Grid.
    axes[0].grid(linestyle='--', linewidth=0.35, alpha=0.5)
    axes[1].grid(linestyle='--', linewidth=0.35, alpha=0.5)
    axes[2].grid(linestyle='--', linewidth=0.35, alpha=0.5)

    # Legend.
    axes[0].legend(loc=0)
    axes[1].legend(loc=0)
    axes[2].legend(loc=0)

    # Tight layout
    plt.tight_layout()

    # Show.
    # plt.show()



###############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
###############################################################################
# Libraries.
import math
import pickle
import inspect
import warnings
import numpy as np
import pandas as pd

from scipy.stats import norm
from sklearn.model_selection import ParameterGrid


class BaseWrapper(object):
    # This is the name of the class.
    _name = 'BASE'

    # Main attributes of the class.
    _raw = None  # The raw object (statsmodels, scipy, ...)
    _result = {}  # Dictionary with metrics filled in self._init_result().
    _config = {}  # Dictionary with configuration filled in
    _conkwargs = {}
    _fitkwargs = {}

    # ---------------------------------------------------------------------------
    #                               INIT METHOD
    # ---------------------------------------------------------------------------
    def __init__(self, **kwargs):
        """Constructor empty defined just so grid_search works in main.

        Note: Emptying the configuration attribute is important because when
              calling grid_search, the self.__class__(args) calls the __init__
              method of the instance and updates the self._config dictionary.
              Since it has not been deepcopied, such modification will alter
              previous wrappers created during the grid search.
        """
        self._raw = None
        self._result = {}
        self._config = {}
        self._conkwargs = {}
        self._fitkwargs = {}

    def _identifier(self):
        """The name to identify this object."""
        return "%s" % self._name

    # ---------------------------------------------------------------------------
    #                        HELPER METHODS (PRIVATE)
    # ---------------------------------------------------------------------------
    def __getattr__(self, name):
        """This method allows to call series attributes from wrapper instance."""
        if name in self._result: return self._result[name]
        if name in self._config: return self._config[name]
        if name in self._conkwargs: return self._conkwargs[name]
        if name in self._fitkwargs: return self._fitkwargs[name]
        raise AttributeError

    def _cast_float(self, e):
        """This method casts an element to float when feasible.
        """
        try:
            return float(e)
        except:
            return e

    def _init_result(self):
        """This method fills self._results with the scores in self._raw.
        """
        pass

    def _init_config(self):
        """This method fills self._config with the configuration."""
        pass

    # ---------------------------------------------------------------------------
    #                              BASIC METHODS
    # ---------------------------------------------------------------------------
    def fargs(self, kwargs, function):
        """This method returns elements in kwargs which are function inputs."""
        # Get all parameters for the function.
        prms = inspect.getfullargspec(function)
        # Create dictionary and return
        return {k: kwargs[k] for k in prms.args if k in kwargs}

    def attrs(self):
        """This method returns all the defined attributes as tuples."""
        return inspect.getmembers(self, lambda a: not inspect.isroutine(a))

    def methods(self):
        """This method returns all the defined methods as tuples."""
        return inspect.getmembers(self, lambda a: inspect.isroutine(a))

    def save(self, fname):
        """This method saves the wrapper."""
        pickle.dump(self.__dict__, open(fname, "wb"))

    def load(self, fname):
        """This method loads the wrapper."""
        self.__dict__.clear()
        self.__dict__.update(pickle.load(open(fname, "rb")))
        return self

        # ---------------------------------------------------------------------------

    #                               GRID SEARCH
    # ---------------------------------------------------------------------------
    def grid_search(self, grid_params):
        """This method computes grid search.

        It creates all the possible combinations combining the arguments passed.
        Such arguments should be for the constructor and the fit methods. The
        results are stored in an array with the corresponding wrapper instances.

        Notes: The wrappers need to have the method fit implemented.

        Parameters
        ----------
        con_kwargs : arguments to be passed to the constructor method.
        fit_kwargs : arguments to be passed to the fit method.

        Returns
        -------
        summary : summary with all the elements.
        """
        # Create empty list.
        grid_results = []

        # Loop for all possible combinations.
        for i, params in enumerate(ParameterGrid(grid_params)):
            try:
                # Create model and fit
                grid_results.append(self.__class__().fit(**params))
            except Exception as e:
                # Throw warning
                msg = "Iteration %s... failed: %s" % (i, e)
                warnings.warn(msg, RuntimeWarning)

        # Return summary.
        return grid_results

    # --------------------------------------------------------------------------
    #                       CREATES SUMMARY DATAFRAMES
    # --------------------------------------------------------------------------
    def from_list_dataframe(self, wrapper_list, **kwargs):
        """This methods creates a dataframe summary from a list.

        Parameters
        ----------
        wrapper_list : list with wrapper objects.
        flabel       : if include the class _name in the index.

        Returns
        -------
        summary : pandas dataframe.
        """
        # Create summary.
        summary = pd.DataFrame()

        # Loop filling the summary.
        for i, wrapper in enumerate(wrapper_list):
            results = wrapper.as_series(**kwargs).rename(i)
            summary = pd.concat([summary, results], axis=1, join='outer')

        # Return
        return summary.T

    def grid_search_dataframe(self, grid_params, **kwargs):
        """This method computes grid search and stores results in a dataframe.

        Parameters
        ----------
        con_kwargs : arguments to be passed to the constructor method.
        fit_kwargs : arguments to be passed to the fit method.

        Returns
        -------
        summary : summary with all the elements.
        """
        # Compute grid search.
        grid_results = self.grid_search(grid_params=grid_params)

        # Create empty dataframe.
        summary = pd.DataFrame()

        # Loop wrappers and fill it.
        for i, wrapper in enumerate(grid_results):
            results = wrapper.as_series(**kwargs).rename(i)
            summary = pd.concat([summary, results], axis=1, join='outer')

        # Return summary.
        return summary

    # ---------------------------------------------------------------------------
    #                                OVERRIDE
    # ---------------------------------------------------------------------------
    def as_series(self, flabel=True, label=None):
        """This method returns a series with all the information.

        Parameters
        ----------
        label : label to concatenate to index name (e.g. trend to label-trend).

        Returns
        -------
        series : pandas series with the results, configuration and model.
        """
        # Concatenate the configuration.
        s = {}
        s.update(self._result)
        s.update(self._config)
        s.update({'model': self._raw})
        s.update({'id': self._identifier()})
        # No label.
        if not flabel: return pd.Series(s)
        # Concat label at the beginning of the index.
        label = self._name if label is None and hasattr(self, '_name') else label
        f = lambda x: "%s-%s" % (label.lower(), x)
        # Return
        return pd.Series(s).rename(index=f, copy=True)

    def as_summary(self):
        """This method displays the final summary."""
        # Call summary function from _raw (if exists).
        fsummary = getattr(self._raw, "summary", None)
        if callable(fsummary):
            return fsummary(self._raw)
        # Show the series information.
        return self.as_series().__repr__()

    def fit(self, **kwargs):
        """This method performs the fit."""
        # Empty results and update configuration.
        self._results = {}
        self._config.update(kwargs)
        return self


if __name__ == '__main__': # pragma: no cover

    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 80)
    pd.set_option('display.precision', 4)

    # Create and fill a base statistic wrapper.
    w = BaseWrapper()
    w._raw = object()
    w._config = {'p': 2, 'c': 4}
    w._result = {'score': 25}

    # Get attributes tuple list.
    print(w.attrs())

    # Get methods tuple list.
    print(w.methods())

    # Get series with parameters.
    print(w.as_series())

    # Print summary.
    print(w.as_summary())

    # Quick access to an attribute.
    print(w.score)

    # -----------
    # Grid search
    # -----------
    # Parameters
    con_params = {
        'con_1': [True],
        'con_2': ['1', '2'],
    }
    fit_params = {
        'fit_1': [5]
    }

    # Grid search method
    summary = w.grid_search_dataframe(con_kwargs=con_params,
                                      fit_kwargs=fit_params)

    # Show
    print(summary)



##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
#
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division 

# Libraries
import sys
import math
import inspect
import warnings
import itertools
import numpy as np
import pandas as pd

# Libraries.
from scipy.stats import norm
from operator import attrgetter

# Import ARIMA from statsmodels.
from statsmodels.tsa.arima_model import ARIMA

# Add module wrappers to sys path dynamically.
sys.path.append("../../../")

# Libraries wrapper.
from pyamr.core.regression.wregression import BaseRegressionWrapper

class PyramidWrapper(BaseRegressionWrapper):

  # Attributes.
  _name = 'PYRAMID'  # Label to add to the attributes when saving.

  def __init__(self, **kwargs):
    """
    """
    # Library.
    from pyramid.arima import ARIMA as PYARIMA
    # Save config parameters.
    super(PyramidWrapper, self).__init__(**kwargs)
    # Create model
    if len(kwargs):
      self._model = PYARIMA(**kwargs)

  def _identifier(self):
    """This method creates a name that describes de model."""
    try:    exogenous = self.exogenous is not None
    except: exogenous = False
    return "name to do"
    #return "%s%sx%s [%s,%s]" % (self._name,
    #                            self.order,
    #                            self.seasonal_order,
    #                            self.trend,
    #                            exogenous) 

  # --------------------------------------------------------------------------
  #                           SET VARIABLES
  # --------------------------------------------------------------------------
  def _params_from_summary(self):
    """Gets parameters from the summary result of the raw object.
    """
    # Format summary
    summary = self._raw.summary().as_csv()
    summary = summary.split("\n",1)[1]     # Remove first line.
    summary = summary.replace("\n",",")    # Replace \n by comma.

    # Split in elements.
    elements = summary.split(",")
    elements = [self._cast_float(e.strip()) for e in elements]

    # Create series.
    d = {}

    # Add parameters.
    d['s_jb_value'] = elements[-13]
    d['s_jb_prob'] = elements[-9]
    d['s_skew'] = elements[-5]
    d['s_Q_value'] = elements[-15]
    d['s_Q_prob'] = elements[-11]
    d['s_H_value'] = elements[-7]
    d['s_H_prob'] = elements[-3]
    d['s_kurtosis'] = elements[-1]
    d['s_heteroskedasticity'] = elements[-7]
    d['s_omnibus_value'] = None
    d['s_omnibus_prob'] = None

    # Return
    return d

  def _init_result(self, alpha=0.05):
    """This method set all the variables into this class.

    @see: statsmodels.Arima
    @see: statsmodels.ArimaResults

    Parameters
    ---------- 
    alpha :

    Returns
    -------
    series :
    """
    # Create series.
    d = {}
    
    # Add generic metrics.
    d['aic'] = self._raw.aic()
    d['aicc'] = self._raw.aicc()
    d['bic'] = self._raw.bic()
    d['hqic'] = self._raw.hqic()
    d['llf'] = self._raw.arima_res_.llf

    # Check if it is arima or sarimax and get corresponding values
    if self._raw.seasonal_order is not None:
      statistic_values = self._raw.arima_res_.zvalues
    else:
      statistic_values = self._raw.arima_res_.tvalues,

    # Create params information.
    params_data = zip(self._raw.arima_res_.data.param_names,
                      self._raw.arima_res_.params,
                      self._raw.arima_res_.bse,
                      statistic_values,
                      self._raw.arima_res_.pvalues,
                      self._raw.arima_res_.conf_int(alpha))

    # Add coefficients statistics to series.
    for name, coef, std, tvalue, pvalue, (cil, ciu) in params_data:
      d['%s_%s'%(name, 'coef')] = coef
      d['%s_%s'%(name, 'std')] = std
      d['%s_%s'%(name, 'tvalue')] = tvalue
      d['%s_%s'%(name, 'tprob')] = pvalue
      d['%s_%s'%(name, 'cil')] = cil
      d['%s_%s'%(name, 'ciu')] = ciu

    # Further statistics
    d.update(self._resid_stats())

    # We cannot use the params_from_summary because this wrappers stores 
    # different models with different summaries. The right way to solve this
    # is by performing the statistics related with the residuals in the
    # regression_wrapper._resid_stats.
    #d.update(self._params_from_summary())

    # Return
    return d

  def _init_config(self):
    """This method initialises the configuration.

    For some reason the interestin data is in the method __init__ for the
    object self._raw (ARIMA) and the method fir for the object
    self._raw.arima_res_.model.

    TODO: Handle if the instances passed to getargspecdict do not exist.
    """
    # Create dictionary.
    d = {}

    # Fill it.
    d.update(self._getargspecdict(self._raw.arima_res_.model, 'fit'))
    d.update(self._getargspecdict(self._raw, '__init__'))

    # Return
    return d

  # --------------------------------------------------------------------------
  #                           HELPER METHODS
  # --------------------------------------------------------------------------
  def as_summary(self, **kwargs):
    """This method displays the summary.
    """ 
    # Elements to split by.
    find = "="*78
    # Split and fill.
    smry = find.join(self._raw.summary(**kwargs).as_text().split(find)[:-1])
    smry = smry.split("\n")
    smry[-6] = smry[-6].replace('=','',5)
    smry[-5:] = [v.replace(' ', '', 5) for v in smry[-5:]]
    smry = "\n".join(smry[:-1])
    # Variables.
    om, omp, dw = 0.0, 0.0, self.m_dw
    jb, jbp = self.m_jb_value, self.m_jb_prob
    nm, nmp = self.m_nm_value, self.m_nm_prob
    skew, kurt = self.m_skew, self.m_kurtosis
    # Add in new lines.
    smry+= "\n%s\n%s\n%s\n" % ("="*78, "Manual".center(78, ' '), "-"*78)
    smry+= "Omnibus:    %#25.3f   Durbin-Watson: %#23.3f\n" % (om, dw)
    smry+= "Normal (N): %#25.3f   Prob(N):       %#23.3f\n" % (nm, nmp) 
    smry+= "="*78 + "\n"
    smry+= "Note that JB, P(JB), skew and kurtosis have different values.\n"
    smry+= "Note that Prob(Q) tests no correlation of residuals."
    # Return
    return smry
    
  # --------------------------------------------------------------------------
  #                                 FIT
  # --------------------------------------------------------------------------
  def fit(self, **kwargs):
    """This method fits the specified arima model.

    Parameters
    ----------
    endog    :
    exog     :
    missing  :
    hasconst :

    Returns
    -------
    object : A PyramidWrapper object.
    """
    # Fill config.
    self._config.update(kwargs)
    # Set model
    self._raw = self._model.fit(**kwargs)
    # Set residuals as attribute.
    self._resid = self._raw.resid()
    # Set series with interesting params.
    self._result = self._init_result(alpha=0.05)
    # return object.
    return self
    
  # ---------------------------------------------------------------------------
  #                               PREDICTION
  # ---------------------------------------------------------------------------
  def get_prediction(self, **kwargs):
    """
    """
    # Compute prediction
    forecast = self._raw.predict_in_sample(**kwargs)
    # Get plotting values.
    mean = forecast.reshape(1,-1)
    cilo = self.conf_int_insample(mean, alpha=0.05)[:,0].reshape(1,-1)
    ciup = self.conf_int_insample(mean, alpha=0.05)[:,1].reshape(1,-1)
    # Time. 
    time = self._time(forecast=mean, **kwargs).reshape(1,-1)
    # Get plotting values.
    return np.concatenate((time, mean, cilo, ciup), axis=0)

  def _time(self, forecast, start=None, **kwargs):
    """This method....
    """
    # Get default start.
    if start is None:
      start = getattr(self._raw.arima_res_, 'k_diff', 0)
    # Return
    return np.arange(forecast.shape[1]) + start

  # ---------------------------------------------------------------------------
  #                               FIND AUTO
  # ---------------------------------------------------------------------------
  def from_instance(self, arima, **kwargs):
    """This method constructs a PyramidWrapper object from pyramid.ARIMA
    """
    # Create object.
    instance = PyramidWrapper()
    # Set model.
    instance._raw = arima
    # Set residuals as attribute.
    instance._resid = arima.resid()
    # Set series with interesting params.
    instance._result = instance._init_result(alpha=0.05)
    # Set configuration parameters.
    instance._config = instance._init_config()
    # Return
    return instance


  def auto(self, **kwargs):
    """This method finds the best arima.

    @see pyrmid.arima.auto_arima

    Parameters
    ----------

    Returns
    -------
    """
    # Library.
    from pyramid.arima import auto_arima
    from pyramid.arima.arima import ARIMA

    # Compute auto_arima.
    results = auto_arima(**kwargs)

    # Return a single PyramidWrapper object.
    if isinstance(results, ARIMA):
      return [self.from_instance(results)]

    # Return an array of PyramidWrapper objects.
    if isinstance(results, list):
      return [PyramidWrapper().from_instance(a) for a in results]








if __name__ == '__main__': # pragma: no cover

  # Libraries.
  import matplotlib.pyplot as plt

  # Set pandas configuration.
  pd.set_option('display.max_colwidth', 14)
  pd.set_option('display.width', 80)
  pd.set_option('display.precision', 4)

  # Constants
  length = 100
  offset = 100
  slope = 10

  # Create time-series.
  x = np.arange(length)
  n = np.random.randn(length)*150
  y = slope*x + offset + n

  # Format exog to be used in pyramid-arima.
  exog = x.reshape(-1,1)

  # Create object
  # This is a wrapper for the methods arima and sarimax within statstools. Take
  # to account the following things when using this wrapper:
  # - seasonal_order = None calls arima, otherwise it calls sarimax.
  # - transparams = True enforces stationarity (not sure).
  # - the exogenous variable needs to be reshaped.
  # - the trend is passed in the constructor (insted of the fit method).
  pyramid = PyramidWrapper(order=(0,1,1), 
                           seasonal_order=(0,0,0,0),
                           trend='c',
                           disp=0).fit(y=y, exogenous=None, fit_args={})

  # Print series.
  print(pyramid.as_series())

  # Print summary.
  print(pyramid.as_summary())

  # -----------------
  # Save & Load
  # -----------------
  # TODO: There is a TypeError: can't pickle zStatespace objects when the
  #       seasonal order is included in the PyramidWrapper. However, 
  #       apparently we are able to save sarimax with seasons using the
  #       SARIMAXWrapper.
  # File location
  #fname = '../../examples/saved/pyramid-sample.pickle'

  # Save
  #pyramid.save(fname=fname)

  # Load
  #pyramid = PyramidWrapper().load(fname=fname)


  # -----------------
  #  Predictions
  # -----------------
  # Variables.
  s, e, d = 10, 110, False

  # Compute predictions (exogenous?).
  preds = pyramid.get_prediction(start=s, end=e, dynamic=d, exogenous=None)

  # Plot truth values.
  plt.plot(y, color='#A6CEE3', alpha=0.5, marker='o',
              markeredgecolor='k', markeredgewidth=0.5,
              markersize=5, linewidth=0.75, label='Observed')

  # Plot forecasted values.
  plt.plot(preds[0,:], preds[1,:], color='#FF0000', alpha=1.00, 
              linewidth=2.0, label=pyramid._identifier())

  # Plot the confidence intervals.
  plt.fill_between(preds[0,:], preds[2,:], 
                               preds[3,:], 
                               color='r', 
                               alpha=0.1)
          
  # Legend 
  plt.legend()

  plt.show()


  # -------------------
  #   GRID SEARCH
  # -------------------
  # Grid parameters.
  fit_params = {'exogenous': [None, exog],
                'y': [y]}
  con_params = {'disp': [0], 
                'order': [(0,1,1)], 
                'seasonal_order': [None],
                'trend': ['n','c','t','ct']}

  # Get summary.
  summary = pyramid.grid_search_dataframe(con_kwargs=con_params, 
                                          fit_kwargs=fit_params,
                                          flabel=False)

  # Plot result.
  print("\n\n")
  print("===================")
  print("Grid Search Summary")
  print("===================")
  print(summary)

  # -------------------
  #    AUTO ARIMA
  # -------------------

  # Example 1: Retrieving just the best fit.
  # ----------------------------------------
  # Find best arima fit.
  stepwise_fit = PyramidWrapper().auto(y=y, trend='c', 
    exogenous=None, 
    start_p=0, start_q=1, max_p=0, max_q=5, d=1,
    seasonal=False, start_P=0, D=1, 
    trace=True,
    error_action='ignore',       # ignore if an order does not work.
    suppress_warnings=True,      # ignore convergence warnings.
    information_criterion='bic', # score to select best model.
    return_valid_fits=False,      # Return all valid fits.
    stepwise=True)               # set to stepwise

  # Get summary.
  summary = PyramidWrapper().from_list_dataframe(stepwise_fit, flabel=False)

  # Print result.
  print("\n\n")
  print("=====================")
  print("Auto Stepwise Summary")
  print("=====================")
  print(summary)


  # Retrieving all fits
  # -------------------
  # Do a grid search using auto_arima.
  brute_fit = PyramidWrapper().auto(y=y, 
    exogenous=None, trend='c',
    start_p=0, d=1, start_q=0, max_p=0, max_d=1, max_q=5, 
    start_P=1, D=None, start_Q=1, max_P=2, max_D=1, max_Q=2, 
    max_order=20, m=1, seasonal=False, stationary=False, 
    information_criterion='bic', alpha=0.05, stepwise=False, 
    return_valid_fits=True,
    test='kpss', seasonal_test='ch', n_jobs=1, start_params=None, method=None, 
    transparams=True, solver='lbfgs', maxiter=50, disp=0, callback=None,
    offset_test_args=None, seasonal_test_args=None, suppress_warnings=False, 
    error_action='ignore', trace=False, random=False, random_state=None, 
    n_fits=None, out_of_sample_size=0, scoring='mse', scoring_args=None)
 
  # Show length
  print(len(brute_fit))

  # Create summary dataframe
  summary = PyramidWrapper().from_list_dataframe(brute_fit, flabel=False)

  # Plot result.
  print("\n\n")
  print("========================")
  print("Auto Brute Force Summary")
  print("========================")
  print(summary)

  # Show.
  plt.show()


###############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
###############################################################################
# Forces decimals on divisions.
from __future__ import division 

# Libraries
import sys
import math
import inspect
import numpy as np
import pandas as pd
import statsmodels.api as sm


# External libraries
from scipy.stats import norm
from sklearn.model_selection import ParameterGrid

# Add module wrappers to sys path dynamically.
sys.path.append("../..")

# Import base wrapper
from pyamr.core.regression.wbase import BaseWrapper


class BaseRegressionWrapper(BaseWrapper):

  # This is the name of the class.
  _name = 'REGRESSION'

  # Main attributes of the class.
  _resid = None

  # ---------------------------------------------------------------------------
  #                             HELPER METHODS
  # ---------------------------------------------------------------------------
  def _init_config(self):
    """This method fills self._config with the configuration."""
    # Create dir.
    d = {}
    # Find attributes values for interesting methods.
    d.update(self._getargspecdict(self._raw.model, '__init__'))
    d.update(self._getargspecdict(self._raw.model, 'fit'))
    # Return
    return d

  def _getargspecdict(self, instance, funcname):
    """This method creates a dictionary with pairs name and value.

    Parameters
    ----------
    instance : object with values
    funcname : function which parameters name will be looked for. 

    Returns
    -------
    tpls : dictionary with argument name and value.
    """
    try:
      # Get argument parameters.
      func = getattr(instance, funcname, None)
      prms = inspect.getargspec(func)
      tpls = {}
      # Create and fill dictionary
      for name in prms.args:
        if name=='self': continue
        tpls[name] = getattr(instance, name, None)
      # Return
      return tpls
    except Exception as e:
      # Print
      print("[Exception at _getargspecdict : %s" % e)
      # Return
      return {}

  # ---------------------------------------------------------------------------
  #                              BASIC METHODS
  # ---------------------------------------------------------------------------
  def _params_from_summary(self):
    """This method returns params from summary.
    """
    return {}

  # ---------------------------------------------------------------------------
  #              STATISTIC METHODS FOR REGRESSION ANALYSIS
  # --------------------------------------------------------------------------- 
  def _resid_stats(self, resid=None, alpha=0.05):
    """This method computes basic stats on the residuals 

    Parameters
    ----------
    resid : array-like
      The residuals to perform the stats on.

    alpha : int-like
      The alpha selected.

    Returns
    -------
    dictionary with the stats for the residuals
    """
    # Check if resid is passed.
    if resid is None:
      resid = self._resid
    
    # No resid to work with.
    if resid is None: 
      return {}

    # Create series.
    d = {}
    
    # Compute autocorrelation (durbin-watson)
    from statsmodels.stats.stattools import durbin_watson
    d['m_dw'] = durbin_watson(resid)

    # Compute normalility (jarque bera).
    from statsmodels.stats.stattools import jarque_bera
    jb_value, jb_prob, skew, kurtosis = jarque_bera(resid)
    d['m_jb_value'] = jb_value
    d['m_jb_prob'] = jb_prob
    d['m_skew'] = skew
    d['m_kurtosis'] = kurtosis

    # Compute normal test (normal test)
    from scipy.stats import normaltest
    nm_value, nm_prob = normaltest(resid)
    d['m_nm_value'] = nm_value
    d['m_nm_prob'] = nm_prob

    # Compute the kolmogorov-smirnov test.
    from scipy.stats import kstest
    ks_value, ks_prob = kstest(resid, 'norm')
    d['m_ks_value'] = ks_value
    d['m_ks_prob'] = ks_prob

    # Compute the shapiro-wilkinson test.
    from scipy.stats import shapiro
    sh_value, sh_prob = shapiro(resid)
    d['m_shp_value'] = sh_value
    d['m_shp_prob'] = sh_prob

    # Compute anderson-darling. 
    # The null hypothesis (sample data is drawn from a population that 
    # follows a particular distribution; in this case normal) can be rejected 
    # if the statistic es larger than the critical values for an specified 
    # significante level.
    from scipy.stats import anderson
    ad_value, ad_cv, ad_sl = anderson(resid)
    d['m_ad_value'] = ad_value
    d['m_ad_nnorm'] = ad_value<ad_cv[2]

    # Return
    return d


  def conf_int_insample(self, forecast, resid=None, alpha=0.05):
    """This function computes a basic confidence interval.

    Note: It might not be the adecuate way of computing it.

    Parameters
    ----------
    forecast : the forecasted values.
    alpha    : the alpha value selected.

    Returns
    -------
    cilo :
    ciup :
    """
    # Check if resid is passed.
    if resid is None:
      resid = self._resid
    
    # No resid to work with.
    if resid is None: 
      return np.array([[],[]])

    # Compute variables.
    const = norm.ppf(1.0-alpha/2.0)
    mu = np.mean(resid)
    std = np.std(resid)
    c = const*(std/math.sqrt(resid.shape[0]))

    # Compute confidence interval.
    cilo = (forecast - c).reshape(-1,1)
    ciup = (forecast + c).reshape(-1,1)

    # Return
    return np.concatenate((cilo, ciup), axis=1)
   

  def _exog(self, start=None, end=None):
    """This method generates the exogenous variable time. 

    Note: it is only used for those regression methods that do not support
    the parameters stard and end in the prediction (wls, ols, rlm, ...). 
    On the other side, this method is not necessary for ARIMA since they
    already support this notation. 

    .. note:: end is included (see _time()).

    Parameters
    ----------
    start : int (optional)
      The time t to start the prediction

    end : int (optional)
      The time t to end the prediction

    Returns
    -------
    the exogenous variable
    """
    # Default start and end.
    exog = self._time(start=start, end=end)
    # Add constant if required.
    trend = getattr(self, 'trend', None)
    # Return exog without constant
    if trend is None or trend=='ct':
      return exog
    # Add constant.
    exog = sm.add_constant(exog)
    # Return
    return exog


  def _time(self, start=None, end=None):
    """This method generates the time variable.

    Note: The value indicated by 'end' is included. This is that way 
    to match with the implementation of ARIMA from statsmodels.

    Parameters
    ----------
    start : int (optional)
      The time t to start the prediction

    end : int (optional)
      The time t to finish the prediction.

    Returns
    -------
    """
    start = 0 if start is None else start
    end = len(self.endog) if end is None else end+1 
    # Generate time variable.
    return np.arange(start, end, 1)

    
  # ---------------------------------------------------------------------------
  #                           BASIC PLOT METHODS
  # ---------------------------------------------------------------------------
  # TODO: Might be useful to create some basic plot methods so it is easy
  # to visualize the results. All the regression wrappers have an example
  # of such basic plotting in their __main__.




if __name__ == '__main__':

  # Constants
  length = 100
  offset = 100
  slope = 10

  # Create time-series.
  x = np.arange(length)
  n = np.random.randn(length)*10
  y_orig = slope*x + offset + n
  y_pred = slope*x + offset

  # Create and fill a base statistic wrapper.
  w = BaseRegressionWrapper()
  w._resid = y_orig - y_pred
  
  # Print resid stats
  print(pd.Series(w._resid_stats()))

  # Print resid confidence intervals
  print(w.conf_int_insample(forecast=y_pred))


# Libraries
import warnings
import pandas as pd

def dri(*args, **kwargs):
    return drug_resistance_index(*args, **kwargs)

def drug_resistance_index_v2(smmry, cu='use', cr='sari',
                          return_all=False,
                          reference_time=None,
                          **kwargs):
        """Computes the Drug Resistance Index

        An possible summary table would look like this...

        DATE     MICROORGANISM ANTIMICROBIAL     sari   use
        2011 Q2  E. Coli       Aminopenicillins  0.422  300
        2011 Q2  E. Coli       Quinolones        0.130  250
        2011 Q2  E. Coli       Cephalosporins    0.010  100
        2011 Q3  E. Coli       Aminopenicillins  0.437  250
        2011 Q3  E. Coli       Quinolones        0.132  300
        2011 Q3  E. Coli       Cephalosporins    0.014  1500


        Parameters
        ----------
        smmry: pd.DataFrame
            The summary DataFrame with the data required to compute the
            drug resistance index. The following information needs to be
            present in the DataFrame:
               (i) the date (e.g. DATE)
               (ii) the resistance (e.g. sari)
               (iii) the drug use (e.g. use)

        cu: str
            Column name with use
        cr: str
            Column name with resistance
        ct: str
            Column name with time
        **kwargs
            Arguments to pass to groupby

        Returns
        -------
        """

        # Enable to chose whether to return all columns or only dri.
        # Ensure that the summary matrix is consistent

        # Clone matrix
        m = smmry.copy(deep=True)

        # Compute
        m['use_period'] = m \
            .groupby(**kwargs)[cu] \
            .transform(lambda x: x.sum())
        m['u_weight'] = (m[cu] / m.use_period)  # .round(decimals=2)
        m['w_rate'] = (m[cr] * m.u_weight)      # .round(decimals=3)
        m['dri'] = m \
            .groupby(**kwargs).w_rate \
            .transform(lambda x: x.sum())

        # Check result for validity.
        #if (m.dri > 1).any():
        #    raise warnings.warn("""
        #            The dri column is ill defined because it has
        ##            values larger than one. Please revisit the
        #            summary table and ensure that all the data
        #            is consistent with the requirements.""")

        """
        if reference_time is not None:
            for t in reference_time:
                # Get use_period uses
                aux = m.groupby(**kwargs).use_period.first()

                use = aux.values[0]
                u_weight = (m[cu] / use)
                w_rate = (m[cr] * u_weight)
                print(w_rate)

                a = m.groupby(**kwargs).groups.keys()
                print(a)
                m['dri_%s' % t] = m \
                    .groupby(**kwargs).w_rate1 \
                    .transform(lambda x: x.sum())

        print(m)
        """

        if return_all:
            return m

        # Update use
        m.use = m.groupby(**kwargs).use \
            .transform(lambda x: x.sum())
        return m.drop(columns=[
            'use_period', 'u_weight', 'w_rate']) \
                .groupby(**kwargs).first()


def drug_resistance_index(dataframe,
            return_complete=False,
            return_usage=False):
    """Computes the DRI.

    .. note:

    Parameters
    ----------

    Returns
    -------
    """
    # Required columns
    required = ['USE', 'RESISTANCE']

    # Check columns
    if set(required).difference(dataframe.columns):
        raise ValueError("The following columns are missing: {0} " \
            .format(set(required).difference(dataframe.columns)))

    # Clone matrix
    m = dataframe.copy(deep=True)

    # Compute
    u, r = m.USE, m.RESISTANCE
    wu = u / u.sum()
    wr = r * wu

    # Return
    if return_complete:
        m['use_period'] = u.sum()
        m['u_weight'] = wu
        m['w_rate'] = wr
        m['dri'] = wr.sum()
        return m
    if return_usage:
        return pd.Series({
            'use_period': u.sum(),
            'dri': wr.sum()
        })
    return wr.sum()



class DRI:
    """

    """
    # Attributes
    c_spe = 'SPECIMEN'
    c_org = 'MICROORGANISM'
    c_abx = 'ANTIMICROBIAL'
    c_dat = 'DATE'
    c_out = 'SENSITIVITY'
    c_drg = 'DRUG'

    c_res = 'RESISTANCE'
    c_use = 'USE'


    """"""
    def __init__(self, column_resistance=c_res,
                       column_usage=c_use):
        """"""
        # Create dictionary to rename columns
        self.rename = {
            column_resistance: self.c_res,
            column_usage: self.c_use
        }

    def compute(self, dataframe, groupby=None,
                column_usage=None, **kwargs):
        """Computes the DRI index.

        .. note:: Include checks.

        Parameters
        ----------
        dataframe: pd.DataFrame
            The pandas dataframe with the information. The following columns
            are always required [RESISTANCE and USE]. The RESISTANCE columns
            indicates the proportion of resistance isolates and the USE the
            amount of antimicrobial doses applied.

        groupby: list, default=None
            The elements to groupby (pd.groupby)

        column_usage: str
            The column with the usage values. This value overwrites the
            column_usage value passed during the instance creation. If
            the value is None, the default value passed during the instance
            creation will be used.

        **kwargs

        Returns
        -------
        """
        # Bad input type
        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError("""
                The instance passed as argument needs to be a pandas
                DataFrame. Instead, a <%s> was found. Please convert 
                the input accordingly.""" % type(dataframe))

        if isinstance(groupby, str):
            groupby = [groupby]

        # Temporal update of usage column
        rename = self.rename.copy()
        if column_usage is not None:
            rename = {k:v for k,v in rename.items()
                if v != self.c_use}
            rename[column_usage] = self.c_use

        # Rename columns
        aux = dataframe.rename(columns=rename, copy=True)

        # Compute overall
        if groupby is None or not groupby:
            return dri(aux, **kwargs)

        # Compute
        scores = aux.groupby(by=groupby) \
            .apply(dri, **kwargs)

        # Format
        if isinstance(scores, pd.Series):
            scores.rename('dri', inplace=True)

        # Return
        return scores








if __name__ == '__main__':

    # Libraries
    import pandas as pd

    from pathlib import Path

    # ----------------------------------
    # Create data
    # ----------------------------------
    # Define susceptibility test records
    susceptibility_records = [
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-04', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],

        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-04', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],

        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],

        ['2021-01-12', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-12', 'URICUL', 'SAUR', 'ACIP', 'intermediate'],
        ['2021-01-13', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-13', 'URICUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-14', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-14', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-15', 'URICUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-15', 'URICUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-16', 'URICUL', 'SAUR', 'ACIP', 'intermediate'],
        ['2021-01-16', 'URICUL', 'SAUR', 'ACIP', 'intermediate'],
    ]

    # Define prescription test records
    prescription_records = [
        ['2021-01-01', 'PATIENT_1', 'AAUG', 150],
        ['2021-01-02', 'PATIENT_1', 'AAUG', 220],
        ['2021-01-03', 'PATIENT_1', 'AAUG', 150],

        ['2021-01-01', 'PATIENT_2', 'AAUG', 250],
        ['2021-01-02', 'PATIENT_2', 'AAUG', 320],
        ['2021-01-03', 'PATIENT_2', 'AAUG', 350],

        ['2021-01-01', 'PATIENT_3', 'ACIP', 450],
        ['2021-01-02', 'PATIENT_3', 'ACIP', 420],
        ['2021-01-03', 'PATIENT_3', 'ACIP', 450],

        ['2021-01-01', 'PATIENT_4', 'ACIP', 50],
        ['2021-01-02', 'PATIENT_4', 'ACIP', 50],
        ['2021-01-03', 'PATIENT_4', 'ACIP', 50],

    ]

    prescription_records = [
        ['2021-01-01', 'PATIENT_1', 'AAUG', 1500],
        ['2021-01-02', 'PATIENT_1', 'AAUG', 2],
        ['2021-01-03', 'PATIENT_1', 'AAUG', 500],

        ['2021-01-01', 'PATIENT_2', 'AAUG', 2000],
        ['2021-01-02', 'PATIENT_2', 'AAUG', 320],
        ['2021-01-03', 'PATIENT_2', 'AAUG', 350],

        ['2021-01-01', 'PATIENT_3', 'ACIP', 2],
        ['2021-01-02', 'PATIENT_3', 'ACIP', 505],
        ['2021-01-03', 'PATIENT_3', 'ACIP', 1124],

        ['2021-01-01', 'PATIENT_4', 'ACIP', 5],
        ['2021-01-02', 'PATIENT_4', 'ACIP', 643],
        ['2021-01-03', 'PATIENT_4', 'ACIP', 2330],

    ]


    # Create DataFrames
    susceptibility = pd.DataFrame(susceptibility_records,
        columns=['DATE',
                 'SPECIMEN',
                 'MICROORGANISM',
                 'ANTIMICROBIAL',
                 'SENSITIVITY'])

    prescriptions = pd.DataFrame(prescription_records,
        columns=['DATE',
                 'PATIENT',
                 'DRUG',
                 'DOSE'])

    # Format dates
    susceptibility.DATE = pd.to_datetime(susceptibility.DATE)
    prescriptions.DATE = pd.to_datetime(prescriptions.DATE)

    # Show
    print("\nSusceptibility records")
    print(susceptibility.head(5))
    print("\nPrescription records")
    print(prescriptions.head(5))


    # .. note:: Uncomment to load the CDDEP example data instead

    # Load default CDDEP sample
    #path = Path('../datasets/cddep')
    #susceptibility = pd.read_csv(path / 'susceptibility.csv')
    #prescriptions = pd.read_csv(path / 'prescriptions.csv')


    # ------------------------
    # Compute summary table
    # ------------------------
    # Libraries
    from pyamr.core.sari import SARI

    # Create sari instance
    sari = SARI(groupby=['DATE',
                         'SPECIMEN',
                         'MICROORGANISM',
                         'ANTIMICROBIAL',
                         'SENSITIVITY'])

    # Compute susceptibility summary table
    smmry1 = sari.compute(susceptibility,
        return_frequencies=False)

    # Compute prescriptions summary table.
    smmry2 = prescriptions \
        .groupby(by=['DATE', 'DRUG']) \
        .DOSE.sum().rename('use')

    # Combine both summary tables
    smmry = smmry1.reset_index().merge(
        smmry2.reset_index(), how='inner',
        left_on=['DATE', 'ANTIMICROBIAL'],
        right_on=['DATE', 'DRUG']
    )

    # Show
    print("\nSummary")
    print(smmry)


    # -------------------------
    # Compute DRI
    # -------------------------
    obj = DRI(
        column_resistance='sari',
        column_usage='use'
    )

    # Compute DRI overall
    dri1 = obj.compute(smmry)

    # Compute DRI
    dri2 = obj.compute(smmry,
        groupby=['SPECIMEN'])

    # Compute DRI
    dri3 = obj.compute(smmry,
        groupby=['MICROORGANISM'])

    # Compute DRI
    dri4 = obj.compute(smmry,
        groupby=['MICROORGANISM', 'ANTIMICROBIAL'])

    # Compute DRI
    dri5 = obj.compute(smmry,
        groupby=['DATE'],
        return_usage=True)

    # Compute DRI
    dri6 = obj.compute(smmry,
        groupby=['DATE', 'MICROORGANISM'],
        return_usage=True)

    # Compute DRI
    dri7 = obj.compute(smmry,
        groupby=['DATE', 'MICROORGANISM', 'ANTIMICROBIAL'],
        return_usage=True,
        return_complete=True)

    # Compute DRI (return all elements of summary table).
    dri8 = obj.compute(smmry,
        groupby=['MICROORGANISM'],
        return_complete=True)

    # Show
    print("\nDRI (1):")
    print(dri1)
    print("\nDRI (2):")
    print(dri2)
    print("\nDRI (3):")
    print(dri3)
    print("\nDRI (4):")
    print(dri4)
    print("\nDRI (5):")
    print(dri5)
    print("\nDRI (6):")
    print(dri6)
    print("\nDRI (7):")
    print(dri7)
    print("\nDRI (8):")
    print(dri8)


    # --------------------------------------------
    # Compute DRI fixed
    # --------------------------------------------
    # Compute prescriptions on t0.
    use_t0 = prescriptions \
        .groupby(by=['DATE', 'DRUG']) \
        .DOSE.sum().rename('use') \
        .to_frame().reset_index() \
        .groupby('DRUG').use.first()

    # Add to summary table
    smmry = smmry.assign(use_t0=smmry.DRUG.map(use_t0))

    # Define groupby
    groupby = [
        'DATE',
        'MICROORGANISM'
    ]

    # Compute DRI
    dri9a = obj.compute(smmry,
        groupby=groupby,
        return_usage=True)

    # Compute DRI using new USE
    dri9b = obj.compute(smmry,
        groupby=groupby,
        return_usage=True,
        column_usage='use_t0')

    #aux = pd.concat([dri9a, dri9b], axis=1)
    aux = dri9a.merge(dri9b,
        left_index=True, right_index=True,
        suffixes=['', '_fixed'])

    # Concatenate (series)
    #aux = pd.concat([
    #    dri9a.rename('dri'),
    #    dri9b.rename('dri_fixed')], axis=1)

    # Show
    print("\n\n")
    print("\nSummary (variable):")
    print(smmry)
    print("\nDRI (9):")
    print(aux)


    # -------------------------------------------------------------------------
    # Testing
    # -------------------------------------------------------------------------

    # ---------------------------------------------------------------------
    # Success
    # ---------------------------------------------------------------------
    # .. note: All this examples should succeed. At the moment the code
    #          breaks if gram is not included. This is because the data
    #          we have created has duplicated values for each gram.
    #          Should we consider this within the ASAI?

    """
    # Compute DRI overall
    dri1 = dri(smmry)

    # Compute DRI
    dri8 = smmry \
        .groupby(by=['SPECIMEN']) \
        .apply(drug_resistance_index_v2)

    # Compute DRI
    dri2 = smmry \
        .groupby(by=['MICROORGANISM']) \
        .apply(drug_resistance_index_v2)

    dri3 = smmry \
        .groupby(by=['MICROORGANISM', 'ANTIMICROBIAL']) \
        .apply(drug_resistance_index_v2,
               return_components=True)

    # Compute DRI
    dri9 = smmry \
        .groupby(by=['DATE']) \
        .apply(drug_resistance_index_v2)

    dri4 = smmry \
        .groupby(by=['DATE', 'MICROORGANISM']) \
        .apply(drug_resistance_index_v2,
               return_components=True)

    dri5 = smmry \
        .groupby(by=['DATE', 'MICROORGANISM', 'ANTIMICROBIAL']) \
        .apply(drug_resistance_index_v2,
               return_components=True)
    """

    # ---------------------------------------------------------------------
    # Errors
    # ---------------------------------------------------------------------
    # .. note: In the examples below, the method acsi is meant to raise
    #          an error either because any of the required missing columns
    #          is missing or because the configuration is not correct.
    print("\n\nHandling errors:")

    # ---------------------------------------------------------------------
    # Warnings
    # ---------------------------------------------------------------------
    # .. note: In the examples below, the method acsi is meant to show a
    #          warning message either no threshold has been specified or
    #          because thresholds have been specified twice.
    print("\n\nShow warnings:")


    """
    # -------------------------
    # Compute resistance index
    # -------------------------
    r = drug_resistance_index(smmry,
        by=['DATE'],
        return_all=True,
        reference_time=[1,2])

    # Show
    print("\nResult:")
    print(r.round(decimals=3))
    """


################################################################################
# Author:
# Date:
# Description:
#
#
#
# Copyright:
#
#
################################################################################
# Import libraries
import pandas as pd

# Import sari
from pyamr.core.sari import sari




# -------------------------------------------------------------------------
#                            helper methods
# -------------------------------------------------------------------------
class MARI:
    """Multiple Antimicrobial Resistance Index"""
    # Attributes
    c_lab = 'LAB_NUMBER'
    c_spe = 'SPECIMEN'
    c_org = 'MICROORGANISM'
    c_dat = 'DATE'
    c_out = 'SENSITIVITY'

    def __init__(self, groupby=[c_spe,
                                c_org,
                                c_lab,
                                c_out]):
        """Constructor.

        Parameters
        ----------
        groupby: list
            The labels of the columns to groupby. The name of
            the columns it should include is as follows:

                [
                    COLUMN_SPECIMEN,
                    COLUMN_MICROORGANISM,
                    COLUMN_LABORATORY_NUMBER,
                    COLUMN_OUTCOME
                ]

        Returns
        --------
        MARI instance
        """
        self.groupby = groupby


    def compute_v1(self, dataframe, shift=None, period=None,
                 cdate=None, return_frequencies=True,
                 return_isolates=True, **kwargs): # pragma: no cover
        """"""
        # Copy DataFrame
        aux = dataframe.copy(deep=True)

        # Create grouper
        grouper = []
        if shift is not None:
            grouper = [pd.Grouper(freq=shift, key=cdate)]
        grouper = grouper + self.groupby

        # Format as datetime
        if cdate is not None:
            aux[cdate] = pd.to_datetime(aux[cdate])

        # ------------------------------------------
        # Frequencies
        # ------------------------------------------
        #
        rn = {
            'mean': 'mari',
            'count': 'freq'
        }

        # Compute frequencies
        isolates = aux.groupby(grouper) \
            .size().unstack().fillna(0)

        # Include mari by isolate
        isolates['mari'] = sari(isolates, **kwargs)

        # Compute frequencies
        if shift is None:

            # Compute mari
            dataframe = isolates.mari \
                .groupby(level=isolates.index.names[:-1]) \
                .agg(['count', 'mean']) \

        else:

            # Create sum and count per shift
            aux = isolates.reset_index() \
                .groupby([cdate] + grouper[1:-2]) \
                .mari.agg(['sum', 'count']) \
                .reset_index().groupby(grouper[1:-2])


            # Compute sums
            sums = aux.rolling(window=period,
                min_periods=1, on=cdate)['sum'].sum()

            # Compute isolate counts
            counts = aux.rolling(window=period,
                min_periods=1, on=cdate)['count'].sum()

            series = (sums/counts).rename('mari')

            dataframe = pd.concat([sums, counts, series], axis=1)

            # Wrong!!
            # Compute sum and count [sum, count]
            #aux = freqs.groupby([cdate] + grouper[2:-1]) \
            #    .mari.agg('mean').reset_index() \
            #    .set_index(cdate).groupby(grouper[2:-1]) \
            #    .mari.rolling(window=period, min_period=1) \
            #    .mean()

        # Rename columns
        dataframe = dataframe.rename(columns=rn)

        # Remove frequencies
        if not return_frequencies:
            dataframe = dataframe['mari']

        # Return
        if return_isolates:
            return dataframe, isolates
        return dataframe


    def rolling(self, dataframe, period, cdate, shift=None):
        """"""
        if shift is None:
            warnings.warn("""
                The input parameter <shift> is None. Thus, the value 
                of the input parameter <period> (%s) has been used."""
                % period)
            shift = period

        # Grouper
        grouper = [pd.Grouper(freq=shift, key=cdate)]
        grouper = grouper + self.groupby[:-2]

        # Compute frequencies
        freqs = dataframe.groupby(grouper) \
            .agg(intermediate=('intermediate', 'sum'),
                 resistant=('resistant', 'sum'),
                 sensitive=('sensitive', 'sum'),
                 n_records=('freq', 'sum'),
                 n_samples=('sari', 'count'),
                 total=('sari', 'sum')) \
            .reset_index() \
            .set_index(cdate).groupby(grouper[1:]) \
            .rolling(window=period, min_periods=1) \
            .sum().fillna(0)

        # Return
        return freqs


    def grouping(self, dataframe, period, cdate):
        """Groups the data.
        """
        # Create grouper
        if hasattr(dataframe[cdate].dt, str(period)):
            grouper = [getattr(dataframe[cdate].dt, period)]
        else:
            grouper = [pd.Grouper(freq=period, key=cdate)]
        grouper = self.groupby[:-2] + grouper

        # Compute
        freqs = dataframe.groupby(grouper) \
            .agg(intermediate=('intermediate', 'sum'),
                 resistant=('resistant', 'sum'),
                 sensitive=('sensitive', 'sum'),
                 n_records=('freq', 'sum'),
                 n_samples=('sari', 'count'),
                 total=('sari', 'sum'))

        # Return
        return freqs


    def compute_v2(self, dataframe, shift=None, period=None,
                         cdate=None, return_frequencies=True,
                         return_isolates=True, **kwargs):
        """Compute MARI v2.

        .. note: No need to copy because SARI does it for us
        """
        # Libraries
        from pyamr.core.sari import SARI

        # Not allowing period to be a number. The main reason is that the
        # most common interpretation is that scenarios with shift=1D
        # period=2D and shift=1D period=2 should be the same. However, the
        # results are actually different. Because period=2 in rolling will
        # use two adjacent rows without considering time. This introduces
        # inconsistencies where there are time gaps without data.
        if period is not None:
            if not isinstance(period, str):
                raise ValueError("""
                    The input parameter <period> cannot be of %s. Ensure 
                    it is either None or a valid string such as 2D or year.
                    """ % type(period))

        # Compute frequencies
        if period is None and shift is None:

            # Compute freqs from sari
            freqs = SARI(groupby=self.groupby) \
                .compute(dataframe,  **kwargs) \
                .groupby(level=[0,1]) \
                .agg(intermediate=('intermediate', 'sum'),
                     resistant=('resistant', 'sum'),
                     sensitive=('sensitive', 'sum'),
                     n_records=('freq', 'sum'),
                     n_samples=('sari', 'count'),
                     total=('sari', 'sum'))

        else:
            # Format as datetime
            dataframe[cdate] = pd.to_datetime(dataframe[cdate])

            # Define new group by with date
            groupby = self.groupby.copy()
            groupby.insert(3, cdate)

            # Compute mari (sari per sample)
            iso = SARI(groupby=groupby) \
                .compute(dataframe, **kwargs) \
                .reset_index()

            if shift is None:
                freqs = self.grouping(dataframe=iso,
                                      period=period,
                                      cdate=cdate)
            else:
                freqs = self.rolling(dataframe=iso,
                                     period=period,
                                     shift=shift,
                                     cdate=cdate)

        # Add mari
        freqs['mari'] = freqs.total / freqs.n_samples

        # Remove frequencies
        if not return_frequencies:
            freqs = freqs['mari']

        # Return
        if return_isolates:
            return freqs, pd.DataFrame()
        return freqs



    def compute_v3(self, dataframe, shift=None, period=None,
                         cdate=None, return_frequencies=True,
                         return_isolates=True, **kwargs): # pragma: no cover
        """Compute MARI v2.

        .. note: It does not work properly.

        .. note: It is considerably slower. Possibly because it
                 is executing the rolling method twice. Try to
                 remove the first rolling as it is not needed.
        """
        # Libraries
        from pyamr.core.sari import SARI

        # Format as datetime
        if cdate is not None:
            dataframe[cdate] = pd.to_datetime(dataframe[cdate])


        # Create object
        sari = SARI(groupby=['SPECIMEN',
                             'MICROORGANISM',
                             'LAB_NUMBER',
                             'DATE',
                             'SENSITIVITY'])

        # Compute mari (sari per sample)
        isolates = sari.compute(dataframe,
            #shift=shift, period=period, cdate=cdate,
            return_frequencies=return_frequencies, **kwargs)

        """
        from pyamr.core.sari import sari
        # Compute frequencies
        isolates = dataframe.groupby(['SPECIMEN',
                             'MICROORGANISM',
                             'LAB_NUMBER',
                             'DATE',
                             'SENSITIVITY']) \
            .size().unstack().fillna(0)

        # Include mari which is the SARI per isolate
        isolates['sari'] = sari(isolates, **kwargs)
        print(isolates)
        """

        by = ['SPECIMEN', 'MICROORGANISM']
        if cdate is not None:
            by = by + [cdate]

        aux = isolates.reset_index() \
            .groupby(by) \
            .agg(intermediate=('intermediate', 'sum'),
                 resistant=('resistant', 'sum'),
                 sensitive=('sensitive', 'sum'),
                 #n_records=('freq', 'sum'),
                 n_samples=('sari', 'count'),
                 total=('sari', 'sum'))

        if period is not None:
            # Compute rolling
            aux = aux \
                .reset_index() \
                .groupby(by[:2]) \
                .rolling(window=period,
                         min_periods=1, on=cdate).agg('sum')

            # Format result.
            aux.index = aux.index.droplevel(2)
            aux = aux.reset_index() \
                .set_index(by)

        # Add mari
        aux['mari'] = aux.total / aux.n_samples

        # Add number of records
        aux['n_records'] = \
            aux.resistant + \
            aux.sensitive + \
            aux.intermediate

        # Return
        if return_isolates:
            return aux, isolates
        return aux


    def compute_v4(self, dataframe, shift=None, period=None,
                         cdate=None, return_frequencies=True,
                         return_isolates=True, **kwargs): # pragma: no cover
        """Compute MARI v1.
        """

        # Copy DataFrame
        aux = dataframe.copy(deep=True)

        # Warning if dates NaN
        # Warning if elements in groupby any all NaN!

        # Create grouper
        grouper = []
        if shift is not None:
            grouper = [pd.Grouper(freq=shift, key=cdate)]
        grouper = grouper + self.groupby

        # Format as datetime
        if cdate is not None:
            aux[cdate] = pd.to_datetime(aux[cdate])

        # ------------------------------------------
        # Frequencies
        # ------------------------------------------
        # Compute frequencies
        isolates = aux.groupby(grouper) \
            .size().unstack().fillna(0)

        # Include mari which is the SARI per isolate
        isolates['mari'] = sari(isolates, **kwargs)



        # Compute frequencies
        if shift is None:

            # Compute mari
            #dataframe = isolates.mari \
            #    .groupby(level=isolates.index.names[:-1]) \
            #    .agg(['count', 'mean']) \


            dataframe = isolates \
                .groupby(level=isolates.index.names[:-1]) \
                .agg(
                    intermediate=('intermediate', 'sum'),
                    resistant=('resistant', 'sum'),
                    sensitive=('sensitive', 'sum'),
                    n_samples=('mari', 'count'),
                    total=('mari', 'sum')
                )

        else:

            dataframe = isolates.reset_index() \
                .groupby([cdate] + grouper[1:-2]) \
                .agg(intermediate=('intermediate', 'sum'),
                     resistant=('resistant', 'sum'),
                     sensitive=('sensitive', 'sum'),
                     n_samples=('mari', 'count'),
                     total=('mari', 'sum')) \
                .reset_index().groupby(grouper[1:-2]) \
                .rolling(window=period,
                         min_periods=1, on=cdate).agg('sum')

            # Format result.
            dataframe.index = dataframe.index.droplevel(2)
            dataframe = dataframe.reset_index() \
                .set_index(['SPECIMEN',
                            'MICROORGANISM',
                            'DATE'])
        # Add mari
        dataframe['mari'] = dataframe.total / dataframe.n_samples

        # Add number of records
        dataframe['n_records'] = \
            dataframe.resistant + \
            dataframe.sensitive + \
            dataframe.intermediate

        # Remove frequencies
        if not return_frequencies:
            dataframe = dataframe['mari']

        # Return
        if return_isolates:
            return dataframe, isolates
        return dataframe


    def compute(self, dataframe, **kwargs):
        """Compute the Multiple Antimicrobial Resistance Index.

        .. note: The compute_v3 does not work properly

        .. todo: Add parameters to rolling!
        .. todo: Place value at the left, center, right of window?
        .. todo: Ensure that works when time gaps present!
        .. todo: Carefull with various indexes!
        .. todo: Warning if dates NaN
        .. todo: Warning if elements in groupby any all NaN!
        .. todo: Warning if not all samples have been tested with same antimicrobials

        Parameters
        ----------
        dataframe: pd.DataFrame
            A DataFrame with the susceptibility test interpretations
            as columns. The default strategies used (see below) expect
            the following columns ['sensitive', 'intermediate', 'resistant']
            and if they do not appear they weill be set to zeros.

        shift: str
            Frequency value to pass to pd.Grouper.

        period: str, int
            Window value to pass to pd.rolling.

        cdate: string, default=None
            The column that will be used as date.

        return_frequencies: boolean, default=True
            Whether to return the frequencies (isolates) or just the resistance index.

        return_isolates: boolean, default=True
            Whether to return the resistance index for each individual isolate.

        strategy: string or func, default='hard'
            The method used to compute sari. The possible options
            are 'soft', 'medium' and 'hard'. In addition, a function
            with the following signature func(dataframe, **kwargs)
            can be passed.

                (i) ``soft``   as R / R+I+S
                (ii) ``medium`` as R / R+S
                (iii) ``hard``  as R+I / R+I+S
                (iv) ``other``  as R+0.5I / R+0.5I+S [Not yet]

        **kwargs: arguments to pass the strategy function.

        Returns
        -------
        dataframe: pd.Series or pd.DataFrame
            The resistance index (pd.Series) or a pd.Dataframe with the
            resistance index (sari), the sums and the frequencies.

        isolates: pd.DataFrame
            The resistance index and each of the sensitivity value
            counts for each individual isolate.

        """
        #self.compute_v1(dataframe, **kwargs)
        #self.compute_v2(dataframe, **kwargs)
        #self.compute_v3(dataframe, **kwargs)
        return self.compute_v2(dataframe, **kwargs)


if __name__ == '__main__': # pragma: no cover

    # Import libraries
    import time
    import warnings
    import pandas as pd

    from timeit import default_timer as timer

    # Import specific libraries
    from pyamr.core.mari import MARI

    # Filter user warning
    warnings.filterwarnings("ignore", category=UserWarning)



    # ---------------------
    # Create data
    # ---------------------
    # Load data
    data = pd.read_csv('../fixtures/indexes/fixture_mari.csv')


    # Create MARI instance
    mari = MARI(groupby=['SPECIMEN',
                         'MICROORGANISM',
                         'LAB_NUMBER',
                         'SENSITIVITY'])

    # Compute MARI overall
    mari_overall, isolates = mari.compute(data,
        return_frequencies=True,
        return_isolates=True)

    # Compute SARI temporal (ITI)
    mari_iti_1d_1d = mari.compute(data, shift='1D',
        period='1D', cdate='DATE',
        return_isolates=False)

    mari_iti_2d_2d = mari.compute(data, shift='2D',
        period='2D', cdate='DATE',
        return_isolates=False)

    mari_iti_year = mari.compute(data,
        period='year', cdate='DATE',
        return_isolates=False)

    # Compute MARI temporal (OTI)
    mari_oti_1d_2d = mari.compute(data, shift='1D',
        period='2D', cdate='DATE',
        return_isolates=False)

    mari_oti_2d_4d = mari.compute(data, shift='2D',
        period='4D', cdate='DATE',
        return_isolates=False)

    # Show
    print("\nIsolates:")
    print(isolates)
    print("\n\n\nMARI (overall):")
    print(mari_overall)
    print("\n\n\nMARI (iti) | 1D_1D:")
    print(mari_iti_1d_1d)
    print("\n\n\nMARI (iti) | 2D_2D:")
    print(mari_iti_2d_2d)
    print("\n\n\nMARI (iti) | year:")
    print(mari_iti_year)
    print("\n\n\nMARI (oti) | 1D_2D:")
    print(mari_oti_1d_2d)
    print("\n\n\nMARI (oti) | 2D_4D:")
    print(mari_oti_2d_4d)


    # -----------------------
    # Full test
    # -----------------------
    from itertools import product

    # Define possible values
    values = [2, '2D', None, 'year']
    combos = list(product(values, values))

    # Show
    print("\n\nCombinations of params <shift> and <period>:")

    # Loop
    for i, (shift, period) in enumerate(combos):
        print("%2s/%2s. Computing... shift=%-5s | period=%-5s ==> " % \
              (i + 1, len(combos), shift, period), end="")
        try:
            s00, s01 = mari.compute(data, shift=shift, period=period, cdate='DATE')
            s10, s11 = mari.compute_v2(data, shift=shift, period=period, cdate='DATE')
            s20, s21 = mari.compute_v3(data, shift=shift, period=period, cdate='DATE')
            print("Ok! equals_1=%s equals_2=%s" %
                  (s00.equals(s10), s10.equals(s20)))
            #print(s00)
            #print(s10)
            #print(s20)
            #print("\n\n" + "=" * 80)
        except Exception as e:
            print(e)


    # ---------------------------------------------------
    # Test timings
    # ---------------------------------------------------
    # Libraries
    from pyamr.datasets.load import make_susceptibility

    # Constants
    rename = {
        'date_received': 'DATE',
        'laboratory_number': 'LAB_NUMBER',
        'specimen_code': 'SPECIMEN',
        'microorganism_code': 'MICROORGANISM',
        'antimicrobial_code': 'ANTIMICROBIAL',
        'sensitivity': 'SENSITIVITY'
    }

    # Load data
    data = make_susceptibility()
    data = data.rename(columns=rename)

    print("\n\nComparing execution times:")

    # Example 1
    # =========
    t0 = timer()
    mari_overall, isolates = mari.compute(data,
        return_frequencies=True,
        return_isolates=True)
    t1 = timer()
    mari_overall, isolates = mari.compute_v2(data,
        return_frequencies=True,
        return_isolates=True)
    t2 = timer()
    mari_overall, isolates = mari.compute_v3(data,
        return_frequencies=True,
        return_isolates=True)
    t3 = timer()
    print("%.10f | %.10f | %.10f " % (t1-t0, t2-t1, t3-t2))


    # Example 2
    # =========
    t0 = timer()
    mari_iti = mari.compute(data, shift='1D',
        period='1D', cdate='DATE',
        return_isolates=False)
    t1 = timer()
    mari_iti = mari.compute_v2(data, shift='1D',
        period='1D', cdate='DATE',
        return_isolates=False)
    t2 = timer()
    mari_iti = mari.compute_v3(data, shift='1D',
        period='1D', cdate='DATE',
        return_isolates=False)
    t3 = timer()
    print("%.10f | %.10f | %.10f " % (t1-t0, t2-t1, t3-t2))

    # Example 3
    # =========
    t0 = timer()
    mari_oti = mari.compute(data, shift='1D',
        period='2D', cdate='DATE',
        return_isolates=False)
    t1 = timer()
    mari_oti = mari.compute_v2(data, shift='1D',
        period='2D', cdate='DATE',
        return_isolates=False)
    t2 = timer()
    mari_oti = mari.compute_v3(data, shift='1D',
        period='2D', cdate='DATE',
        return_isolates=False)
    t3 = timer()
    print("%.10f | %.10f | %.10f " % (t1-t0, t2-t1, t3-t2))






##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division 

# Libraries
import sys
import numpy as np
import pandas as pd

# Add module wrappers to sys path dynamically.
sys.path.append("../..")

# Import base wrapper
from pyamr.core.stats.wbase import BaseWrapper


class CorrelationWrapper(BaseWrapper):

  # --------------------------------------------------------------------------
  #                          overriden methods
  # --------------------------------------------------------------------------
  def evaluate(self, alpha=0.05):
    """This method set all the variables into this class.
    """
    # Create series.
    d = {}
    # Add results
    d['spearman_corr'] = self._raw['scipy.stats.spearmanr'].correlation
    d['spearman_pval'] = self._raw['scipy.stats.spearmanr'].pvalue
    d['pearson_corr'] = self._raw['scipy.stats.pearsonr'][0]
    d['pearson_pval'] = self._raw['scipy.stats.pearsonr'][1]
    d['crosscorr'] = self._raw['numpy.correlate']
    # Return
    return d
    
  def as_summary(self, alpha=0.05):
    """This method displays the summary.
    """
    # Create summary base
    summary = "         Correlation\n"
    summary+= "==============================\n"
    summary+= "Pearson:           %#11.3f\n" % self.pearson_corr
    summary+= "Spearman:          %#11.3f\n" % self.spearman_corr
    summary+= "Cross correlation: %#11.3f\n" % self.crosscorr
    summary+= "=============================="
    # Return
    return summary

  def fit(self, x1, x2, **kwargs):
    """This method computes kendall for monotonic increase

    NOTES: - scipy pvalues are not reliavble if less than 500 observations.

    Parameters
    ----------
    x1        :
    y2        :

    Returns
    -------
    object : A CorrelationWrapper objects.
    """
    # Library.
    import scipy as sp

    # Empty the class
    self._empty()

    # Update the configuration
    self._config.update(kwargs)

    # Initialize raw data
    self._raw = {} 

    # Compute correlations
    self._raw['numpy.correlate'] = np.correlate(x1, x2)
    self._raw['numpy.corrcoef'] = np.corrcoef(x1, x2)
    self._raw['scipy.stats.pearsonr'] = sp.stats.pearsonr(x1, x2)  
    self._raw['scipy.stats.spearmanr'] = sp.stats.spearmanr(x1, x2) 

    # Evaluate the model
    if self.evaluate:
      self._result = self.evaluate()

    # Save results.
    return self
    


if __name__ == '__main__':
 
  # ----------------------------
  # set basic configuration
  # ----------------------------
  # Set pandas configuration.
  pd.set_option('display.max_colwidth', 80)
  pd.set_option('display.width', 150)
  pd.set_option('display.precision', 4)

  # ----------------------------
  # create data
  # ----------------------------
  # Constants
  length = 100
  offset = 100
  slope = 10

  # Create time-series.
  x = np.arange(length)
  n = np.random.rand(length)
  y1 = n*slope + offset
  y2 = n*(-slope) + offset 

  # ----------------------------
  # create correlation object
  # ----------------------------
  # Create object
  correlation = CorrelationWrapper().fit(x1=y1, x2=y2)

  # Print series.
  print("\n")
  print(correlation.as_series())

  # Print summary.
  print("\n")
  print(correlation.as_summary())

  correlation

  # -----------------
  # Save and load
  # -----------------
  # File location
  #fname = '../examples/saved/correlation-sample.pickle'

  # Save
  #correlation.save(fname=fname)

  # Load
  #correlation = CorrelationWrapper().load(fname=fname)


# Libraries
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib as mpl

from matplotlib import colors


def create_mapper(dataframe, column_key, column_value):
  """This method constructs a mapper

  Parameters
  ----------
  dataframe: dataframe-like
    The dataframe from which the columns are extracted

  column_key: string-like
    The name of the column with the values for the keys of the mapper

  column_value: string-like
    The name of the column with the values for the values of the mapper

  Returns
  -------
  dictionary
  """
  dataframe = dataframe[[column_key, column_value]]
  dataframe = dataframe.drop_duplicates()
  return dict(zip(dataframe[column_key], dataframe[column_value]))


# -----------------------------------------------------------------
#                           Methods
# -----------------------------------------------------------------
def scalar_colormap(values, cmap, vmin, vmax):
    """This method creates a colormap based on values.

    Parameters
    ----------
    values : array-like
    The values to create the corresponding colors

    cmap : str
    The colormap

    vmin, vmax : float
    The minimum and maximum possible values

    Returns
    -------
    scalar colormap
    """
    # Create scalar mappable
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
    mapper = mpl.cm.ScalarMappable(norm=norm, cmap=cmap)
    # Gete color map
    colormap = sns.color_palette([mapper.to_rgba(i) for i in values])
    # Return
    return colormap


def get_category_colors(index, category, cmap='hls'):
    """Creates the colors for the different elements in
    a given categorical feature vector.

    Parameters
    ----------
    values : array-like
        The vector with the categorical values

    cmap: string-like
        The colormap to use

    default: string-like
        The color to be used for the first value. Note that this
        value needs to appear first on the the sorted list, as such
        it is recommended to set is as _default.

    Returns
    -------
    pd.Series
        List of colors.
    """
    # Get categories
    categories = index.get_level_values(category)
    # Get unique elements
    unique = np.unique(categories)
    # Create the palette
    palette = sns.color_palette(cmap, desat=0.5, n_colors=unique.shape[0])
    # Create mappers from category to color
    mapper = dict(zip(map(str, unique), palette))
    # Create list with colors for each category
    colors = pd.Series(categories, index=index).map(mapper)
    # Return
    return colors


def vlinebgplot(ax, top, xv, bg=None):
    """This function adds a vertical line and background

    Parameters
    ----------
    ax: matplotlib axes
    top: float
        The max y value
    xv: float
        The x value
    bg: boolean
        Whether to include a background

    Returns
    -------
    """
    # Libraries
    import seaborn as sns

    # Plot line.
    ax.plot((xv, xv), (-1, top), color='gray',
            linestyle='--', linewidth=1, zorder=0)

    # Plot background.
    if bg is not None:
        cb = sns.color_palette("Set2", 10)[1]
        ax.fill_between([xv, bg], [-1, -1], [top, top],
                        zorder=0, alpha=0.1, color=cb)


def hlinebgplot(ax, right, yv, bg=None):
    """This function adds a vertical line and background

    Parameters
    ----------

    Returns
    -------
    """
    # Libraries
    import seaborn as sns

    # Plot line
    ax.plot((-1, right), (yv, yv), color='gray',
            linestyle='--', linewidth=1, zorder=0)

    # Plot background.
    if bg is not None:
        cb = sns.color_palette("Set2", 10)[1]
        ax.fill_between([-1, right], [yv, yv], [bg, bg],
                        zorder=0, alpha=0.2, color=cb)


def plot_sns_heatmap():
    pass


def plot_sns_clustermap():
    pass


def plot_sns_relmap():
    pass


def plot_spectrum():
    pass


def plot_timeseries():
    pass


class MidpointNormalize(colors.Normalize):
    """Normalise the colorbar so that diverging bars
       work there way either side from a prescribed
       midpoint value)

    Example
    -------
        > MidpointNormalize(midpoint=0., vmin=-100, vmax=100)

    """

    def __init__(self, vmin=None,
                       vmax=None,
                       midpoint=None,
                       clip=False):
        """Constructor
        """
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        """Call

        .. note: Ignoring masked values and all kind of edge cases
                 to keep it simple.
        """
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin,
                self.midpoint,
                self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y), np.isnan(value))


##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division

# Libraries
import sys
import numpy as np
import pandas as pd
import statsmodels.api as sm

# Import pyamr
from pyamr.core.regression.wregression import BaseRegressionWrapper


class TheilSensWrapper(BaseRegressionWrapper):
    # Attributes.
    _name = 'THEILSENS'  # Label to add to the attributes when saving.

    # --------------------------------------------------------------------------
    #                           HELPER METHODS
    # --------------------------------------------------------------------------
    def _init_result(self, alpha=0.05):
        """This method set all the variables into this class.
        """
        # Create series
        d = {}
        # Set results.
        d['slope'] = self._raw[0]
        d['intercept'] = self._raw[1]
        d['ci_lower'] = self._raw[2]
        d['ci_upper'] = self._raw[3]
        # Return
        return d

    def as_summary(self, alpha=0.05):
        """This method displays the summary.
        """
        # Create summary base
        summary = '         TheilSens Slope   \n'
        summary += "==================================\n"
        summary += "slope:     %23s\n" % str(round(self.slope, 4))
        summary += "intercept: %23s\n" % str(round(self.intercept, 4))
        summary += "ci_lower:  %23s\n" % str(round(self.ci_lower, 4))
        summary += "ci_upper:  %23s\n" % str(round(self.ci_upper, 4))
        summary += "=================================="
        # Return
        return summary

    # --------------------------------------------------------------------------
    #                                 FIT
    # --------------------------------------------------------------------------
    def fit(self, **kwargs):
        """This method....

        Parameters
        ----------
        x :
        y :
        alpha :

        Returns
        -------
        object : A KendallWrapper objects.
        """
        # Library.
        from scipy.stats.mstats import theilslopes
        # Save parameters.
        self._config.update(kwargs)
        self._conkwargs = self.fargs(self._config, theilslopes)
        # Compute theilsens slopes
        self._raw = theilslopes(**self._conkwargs)
        # Set series.
        self._result = self._init_result()
        # Compute residuals.
        x, y = kwargs['x'], kwargs['y']
        self._resid = y - (x * self.slope + self.intercept)
        # return
        return self

    def get_prediction(self, exog=None, start=None, end=None, **kwargs):
        """This method...

        Parameters
        ----------
        exog      :
        start     :
        end       :
        kwargs    :

        Returns
        -------
        """
        # Create exogenous variable.
        if start is not None or end is not None:
            exog = self._exog(start, end)

        # Compute prediction
        forecast = exog * self.slope + self.intercept
        # Get plotting values.
        mean = forecast.reshape(1, -1)
        cilo = self.conf_int_insample(mean, alpha=0.05)[:, 0].reshape(1, -1)
        ciup = self.conf_int_insample(mean, alpha=0.05)[:, 1].reshape(1, -1)
        # Add time.
        time = exog.reshape(1, -1)
        # Get plotting values.
        return np.concatenate((time, mean, cilo, ciup), axis=0)


if __name__ == '__main__': # pragma: no cover
    # Libraries.
    import matplotlib.pyplot as plt

    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 150)
    pd.set_option('display.precision', 4)

    # Constants
    length = 100
    offset = 100
    slope = 10

    # Create timeseries.
    x = np.arange(length)
    y = np.random.rand(length) * slope + offset + x

    # Create object
    theilsens = TheilSensWrapper().fit(x=x, y=y)

    # Print series.
    print("\n")
    print(theilsens.as_series())
    # Print summary.
    print("\n")
    print(theilsens.as_summary())

    import sys
    sys.exit()

    # -----------------
    # Save & Load
    # -----------------
    # File location
    fname = '../../examples/saved/theilsens-sample.pickle'

    # Save
    theilsens.save(fname=fname)

    # Load
    theilsens = TheilSensWrapper().load(fname=fname)

    # -----------------
    #  Predictions
    # -----------------
    # Variables.
    start, end, = 10, 70

    # Compute predictions.
    preds = theilsens.get_prediction(start=start, end=end)

    # Plot truth values.
    plt.plot(x, y, color='#A6CEE3', alpha=0.5, marker='o',
             markeredgecolor='k', markeredgewidth=0.5,
             markersize=5, linewidth=0.75, label='Observed')

    # Plot forecasted values.
    plt.plot(preds[0, :], preds[1, :], color='#FF0000', alpha=1.00,
             linewidth=2.0, label=theilsens._identifier())

    # Plot the confidence intervals.
    plt.fill_between(preds[0, :], preds[2, :],
                     preds[3, :],
                     color='r',
                     alpha=0.1)

    # Legend
    plt.legend()

    # Grid search
    # -----------
    # Grid parameters.
    grid_params = {'x': [x], 'y': [y], 'alpha': [0.05, 0.1]}

    # Get summary.
    summary = TheilSensWrapper().grid_search_dataframe(grid_params=grid_params)

    # Plot result (drop x which is an array to improve visualization).
    print(summary)

    # Show
    # plt.show()



"""

Running only one test:

    ​​pytest​​ ​​-v​​ ​​test_manual.py::test_dri_cddep

"""

# Libraries
import pytest
import numpy as np
import pandas as pd

from pathlib import Path

# Specific
from pyamr.core.sari import sari
from pyamr.core.asai import asai
from pyamr.core.sari import SARI
from pyamr.core.asai import ASAI
from pyamr.core.mari import MARI
from pyamr.core.dri import DRI

# ----------------------------------------------------
# Fixtures
# ----------------------------------------------------
# Define path to fixtures folder
fixtures_path = Path(__file__).parent.parent.absolute() / 'fixtures'



@pytest.fixture
def timeseries():
    """"""
    # Constants
    length = 100
    offset = 100
    slope = 10

    # Create time-series.
    np.random.seed(seed=42)
    x = np.arange(length)
    y = np.random.rand(length) * slope + offset
    return x, y

@pytest.fixture
def fixture():
    """This ..."""
    data = [
        [0, 0, 0, np.NaN, np.NaN, np.NaN, np.NaN],
        [0, 0, 1, 0/1, 0.0/1, 0/1, 0/1],
        [0, 1, 0, 1/1, 0.5/1, 0/1, np.NaN],
        [0, 1, 1, 1/2, 0.5/2, 0/2, 0/1],
        [1, 0, 0, 1/1, 1.0/1, 1/1, 1/1],
        [1, 0, 1, 1/2, 1.0/2, 1/2, 1/2],
        [1, 1, 0, 2/2, 1.5/2, 1/2, 1/1],
        [1, 1, 1, 2/3, 1.5/3, 1/3, 1/2]
    ]
    return pd.DataFrame(data=data,
        columns=['resistant', 'intermediate', 'sensitive',
                 'hard', 'medium', 'soft', 'basic'])

def fixture1():
    """This ..."""
    pass

@pytest.fixture
def fixture3():
    return pd.read_csv(fixtures_path / 'fixture_03.csv')

@pytest.fixture
def fixture4():
    return pd.read_csv(fixtures_path / 'fixture_04.csv')

@pytest.fixture
def fixture5():
    return pd.read_csv(fixtures_path / 'fixture_05.csv')

@pytest.fixture
def fixture_index_mari():
    return pd.read_csv(fixtures_path / 'indexes' / 'fixture_mari.csv')


@pytest.fixture
def fixture_dri_cddep_test():
    path = Path(fixtures_path) / 'cddep'
    smmry = pd.read_csv(path / 'summary.csv')
    outpt = pd.read_csv(path / 'outcome.csv')
    return smmry, outpt

@pytest.fixture
def fixture_acsi_lancet_test():
    path = Path(fixtures_path) / 'lancet'
    return pd.read_csv(path / 'mmc2_MIS.csv')



# ------------------------
# Example
# ------------------------
def func(x):
    return x + 1

def test_answer():
    """
    PyTest sample test
    :return:
    """
    assert func(3) == 4


# ----------------------------------------------
#   Single Antibiotic Resistance Index (SARI)
# ----------------------------------------------
def test_sari_method_empty(fixture3):
    assert 4 == 4

def test_sari_empty_dataframe():
    with pytest.raises(Exception) as e:
        sari(pd.DataFrame())

@pytest.mark.parametrize("strategy",
    ['hard', 'medium', 'soft', 'basic',
     pytest.param('other', marks=pytest.mark.xfail)])
def test_sari_strategy(fixture, strategy):
    r = sari(fixture, strategy=strategy)
    assert isinstance(r, pd.Series)
    assert r.equals(fixture[strategy])

def test_sari_class_overall(fixture3):
    r = SARI().compute(fixture3)
    assert isinstance(r, pd.DataFrame)
    assert r.shape[0] == 4

def test_sari_class_timeseries(fixture3):
    r = SARI().compute(fixture3, period='year', cdate='DATE')
    assert isinstance(r, pd.DataFrame)
    assert r.shape[0] == 5
    assert 'DATE' in r.index.names


# ----------------------------------------------
#   Multiple Antibiotic Resistance Index (MARI)
# ----------------------------------------------
def test_mari_class(fixture5):
    r, isolates = MARI().compute(fixture5,
        return_frequencies=True,
        retyrn_isolates=False)
    assert r.shape[0] == 3

def test_mari_class_temporal_oti(fixture5):
    r = MARI().compute(fixture5, shift='1D',
        period='2D', cdate='DATE',
        return_isolates=False)
    assert r.shape[0] == 7

def test_mari_class_temporal_iti_period_is_integer_fails(fixture5):
    with pytest.raises(Exception) as e:
        r = MARI().compute(fixture5, shift='1D',
            period=1, cdate='DATE',
            return_isolates=False)


# ---------------------------------------------------
#   Antimicrobial Spectrum of Activity Index (ASAI)
# ---------------------------------------------------
@pytest.mark.parametrize("columns,kwargs",
    [('RESISTANCE', {}),
     ('GENUS', {}),
     ('SPECIE', {}),
     ('FREQUENCY', {'weights':'frequency'}),
     ('W_GENUS', {'weights':'specified'}),
     ('W_SPECIE', {'weights':'specified'})])
def test_asai_error_missing_column(fixture4, columns, kwargs):
    with pytest.raises(ValueError):
        r = fixture4.drop(columns=columns) \
            .groupby(['ANTIBIOTIC']) \
            .apply(asai, **kwargs)

def test_asai_error_weights_not_valid(fixture4):
    with pytest.raises(ValueError):
        r = fixture4 \
            .groupby(['ANTIBIOTIC']) \
            .apply(asai, weights='invalid')

@pytest.mark.parametrize("column", ['W_GENUS', 'W_SPECIE'])
def test_asai_error_weights_wcolumn_not_valid(fixture4, column):
    with pytest.raises(ValueError):
        aux = fixture4.copy(deep=True)
        aux.loc[0, column] = 1
        r = aux \
            .groupby(['ANTIBIOTIC']) \
            .apply(asai, weights='specified')

def test_asai_warn_threshold_double_defined(fixture4):
    with pytest.warns():
        r = fixture4 \
            .groupby(['ANTIBIOTIC']) \
            .apply(asai, threshold=0.6)

def test_asai_warn_threshold_not_defined(fixture4):
    with pytest.warns():
        r = fixture4.drop(columns=['THRESHOLD']) \
            .groupby(['ANTIBIOTIC']) \
            .apply(asai, threshold=None)

def test_asai_weights_from_column(fixture4):
    r = fixture4 \
        .groupby(['ANTIBIOTIC']) \
        .apply(asai, verbose=0)
    assert r.shape[0] == 2

def test_asai_weights_uniform(fixture4):
    r = fixture4 \
        .groupby(['ANTIBIOTIC', 'GRAM']) \
        .apply(asai, weights='uniform')
    assert r.shape[0] == 4

def test_asai_weights_frequency(fixture4):
    r = fixture4 \
        .groupby(['ANTIBIOTIC', 'GRAM']) \
        .apply(asai, weights='frequency')
    assert r.shape[0] == 4

def test_asai_class(fixture4):
    scores = ASAI().compute(fixture4,
        groupby=['ANTIBIOTIC', 'GRAM'],
        weights='uniform',
        threshold=None,
        min_freq=0)
    assert scores.shape[0]==4


# ---------------------------------------------------
#   Drug Resistance Index (DRI)
# ---------------------------------------------------
def test_dri_cddep(fixture_dri_cddep_test):
    """Validate the output using CDDEP data."""
    summary, output = fixture_dri_cddep_test
    summary['RESISTANCE'] = summary.R / summary.ISOLATES
    r = DRI().compute(summary,
        groupby=['DATE'],
        return_complete=True) \
            .reset_index(drop=True) \
            .round(decimals=3)
    c = ['u_weight', 'w_rate', 'dri']
    assert r[c].equals(output[c])



# ---------------------------------------------------
#   Collateral Resistance Index (CRI)
# ---------------------------------------------------
def test_acsi_lancet(fixture_acsi_lancet_test):
    """
    .. note:: There is a known issue in which the values computed
              for those rows in which the contingency matrix contained
              a 0 is not appropriate.


    """
    # Libraries
    from pyamr.core.acsi import CRI
    from pyamr.core.acsi import mutual_info_matrix_v3

    # Parameters
    rename = {
        'S1S2':'SS',
        'S1R2':'SR',
        'R1S2':'RS',
        'R1R2':'RR'
    }
    cols = list(rename.keys())

    # Assign variable
    data = fixture_acsi_lancet_test
    # Replace zeros with NaN
    data[cols] = data[cols].replace({'0': np.nan, 0: np.nan})
    # Remove rows with NaN
    data = data.dropna(subset=cols+['MIS'], how='any')
    # Compute CRI (need renaming to work)
    data = data.rename(columns=rename)
    data['CRI'] = data.apply(CRI, axis=1,
        args=(mutual_info_matrix_v3,))

    # Check if they are equal
    a = data.MIS.round(decimals=5).to_numpy()
    b = data.CRI.round(decimals=5).to_numpy()
    assert np.array_equal(a, b)



# --------------------------------------
# Statistical tests (statstools)
# --------------------------------------



##############################################################################
# Author: Bernard Hernandez
# Filename: adfuller.py
#
# Description : This file contains a wrapper for the adfuller module.
###############################################################################
# https://mkaz.blog/code/python-string-format-cookbook/
# Future
from __future__ import division

# Libraries
import sys
import numpy as np
import pandas as pd

# Import base wrapper
from pyamr.core.stats.wbase import BaseWrapper


class KPSSWrapper(BaseWrapper):
    """The KPSS wrapper.

    The Kwiatkowski–Phillips–Schmidt–Shin (KPSS) test is used to identify
    whether a time series is stationary around a deterministic trend (thus
    trend stationary) against the alternative of a unit root.

    In the KPSS test, the absence of a unit root is not a proof of stationarity
    but, by design, of trend stationarity. This is an important distinction since
    it is possible for a time series to be non-stationary, have no unit root yet
    be trend-stationary.

    In both, unit-root and trend-stationary processes, the mean can be increasing
    or decreasing over time; however, in the presence of a shock, trend-stationary
    processes revert to this mean tendency in the long run (deterministic trend)
    while unit-root processes have a permanent impact (stochastic trend).

    Notes
    -----
    H0: The series has unit root => Trend Stationary.
    H1: The series has no unit root => No Trend Stationary.

    If p-value > alpha: Failed to reject H0
    If p-value <= alpha: Reject H0
    """
    pass


################################################################################
# Author:
# Date:
# Description:
#
#
#
# Copyright:
#
# 
################################################################################
from __future__ import division

# Libraries
import sys
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib as mpl
import matplotlib.pyplot as plt



# -----------------------------------------------------------------------------
#                               helper methods
# -----------------------------------------------------------------------------
def _get_category_colors(values, cmap='tab20b', default='gray'):
  """This method creates the colors for the different elements in
  categorical feature vector.

  Parameters
  ----------
  values : array-like
    The vector with the categorical values

  cmap: string-like
    The colormap to use

  default: string-like
    The color to be used for the first value. Note that this
    value needs to appear first on the the sorted list, as such
    it is recommended to set is as _default.

  Returns
  -------
  """
  # Get unique elements
  unique = np.unique(values)
  # Sort unique values
  unique.sort()
  # Create the palette (gray for _na)
  palette = [default] + sns.husl_palette(len(unique), s=.45)
  # Create mappers from category to color
  mapper = dict(zip(map(str, unique), palette))
  # Create list with colors for each category.
  colors = pd.Series(values).map(mapper)
  # Return
  return colors


def _idxs(index, level, values):
  """
  """
  return index.get_level_values(level).isin(values)

def _get_colors(index, column, palette):
  """This method gets the colors for each index value.

  # Get species
  vals1, pal1, row_colors = \
    _get_colors(dataframe.index, self.c_gen, 'nipy_spectral')
  vals2, pal2, col_colors = \
    _get_colors(dataframe.columns, self.c_cat, 'tab20b')


  Parameters
  ----------
  index:

  column:

  palette:

  Returns
  -------
  """
  # Get species
  species = index.get_level_values(column).unique().sort_values()
  netwpal = ['gray'] + sns.husl_palette(len(species), s=.45)
  #netwpal = ['gray'] + list(sns.color_palette('Set2', len(species)+1, .75))
  #netwpal = ['gray'] + list(sns.color_palette(palette, len(species)+1, .45))
  netwmap = dict(zip(map(str, species), netwpal))
  # Convert the palette to vectors that will be drawn on the side of the matrix
  networks = index.get_level_values(column)
  network_colors = pd.Series(networks, index=index).map(netwmap)
  # Return
  return species, netwpal, network_colors




class Antibiogram:
    """"""

    # Attributes
    c_cat = 'CATEGORY'
    c_abx = 'ANTIBIOTIC'
    c_org = 'SPECIE'
    c_gen = 'GENUS'
    c_idx = 'SARI'

    def __init__(self, column_organism=c_org,
                     column_antibiotic=c_abx,
                     column_genus=c_gen,
                     column_category=c_cat,
                     column_index=c_idx,
                     colormap_genus='nipy_spectral',
                     colormap_category='tab20b',
                     discard_intrinsic=False):
        """The constructor.

        Parameters
        ----------
        column_organism: string-like
          The column name with the the organism values

        column_antibiotic: string-like
          The column name with the the antibiotic values

        column_genus: string-lik
          The column name with the the organism genus values

        columns_category: string-like
          The column name with the the antibiotic category values

        colormap_genus: string-like
          The colormap to use to display the genus values

        colormap_category: string-like
          The colormap to use to display the category values

        discard_intrinsic: string-like
          Wether or not to remove the intrinsic resistance indexes. These are
          defined as those indexes in which the microbiology was neven found
          to be sensitivity and therefore has a resistance index of 1.

        Returns
        -------
        """
        # Set parameters
        self.colormap_genus = colormap_genus
        self.colormap_category = colormap_category
        self.discard_intrinsic = discard_intrinsic

        # Create dictionary to rename columns
        self.rename_columns = {column_antibiotic: self.c_abx,
                               column_organism: self.c_org,
                               column_genus: self.c_gen,
                               column_category: self.c_cat,
                               column_index: self.c_idx}



    def fit(self, dataframe):
        """This method fits the antibiogram to the data.

        Parameters
        ----------
        dataframe: pd.DataFrame
            The...

        Returns
        -------
        Antibiogram instance
        """
        # Check that it is a dataframe
        if not isinstance(dataframe, pd.DataFrame):
          raise TypeError("The instance passed as argument needs to be a pandas "
                          "DataFrame. Instead, a <%s> was found. Please convert "
                          "the input accordingly." % type(dataframe))

        # Rename columns
        self._dataframe = \
          dataframe.rename(columns=self.rename_columns, copy=True)

        # Fill missing categories (so they have same color)
        self._dataframe[self.c_org].fillna('_na', inplace=True)
        self._dataframe[self.c_cat].fillna('_na', inplace=True)

        # Keep only interesting rows
        self._dataframe = self._dataframe[self.rename_columns.values()]

        # Format dataframe
        self._dataframe = self._dataframe.set_index([self.c_org,
                                                     self.c_abx,
                                                     self.c_gen,
                                                     self.c_cat])

        # Create heatmap
        self._dataframe = self._dataframe.unstack([self.c_abx,
                                                   self.c_cat])

        # Drop sari index
        self._dataframe.columns = self._dataframe.columns.droplevel()

        # Discard intrinsic resistance
        if not self.discard_intrinsic:
          self._dataframe[self._dataframe==1] = np.nan

        # Return
        return self


    def plot(self, organisms=None, antibiotics=None,
                                 genera=None,
                                 categories=None,
                                 method='weighted',
                                 metric='euclidean',
                                 vmin=0.0,
                                 vmax=1.0,
                                 cmap=None,
                                 figsize=(5,5),
                                 xticklabels=True,
                                 yticklabels=True,
                                 **kwargs):
        """This function plots the antibiogram.

        The arguments are those to be passed to the clustermap implementation
        which is available on seaborn.

        Parameters
        ----------
        method: string-like

        metric: string-like

        vmin, vmax: float-like, float-like
          The min and maximum values within the matrix (or feasible range).

        cmap: string-like
          The colormap to be used to plot the resistance indexes.

        figsize: tuple-like
          The size of the figure

        Returns
        -------
        an axis object
        """
        # Define colormap
        if cmap is None:
          cmap = sns.color_palette("Reds", desat=0.5, n_colors=10)

        # Copy dataframe
        dataframe = self._dataframe.copy(deep=True)

        # Filter data
        if organisms is not None:
          dataframe = dataframe.loc[_idxs(dataframe.index, 0, organisms), :]
        if genera is not None:
          dataframe = dataframe.loc[_idxs(dataframe.index, 1, genera), :]
        if antibiotics is not None:
          dataframe = dataframe.loc[:, _idxs(dataframe.columns, 0, antibiotics)]
        if categories is not None:
          dataframe = dataframe.loc[:, _idxs(dataframe.columns, 1, categories)]

        # Create mask
        mask = pd.isnull(dataframe)

        # Fill nan with a very small value
        dataframe = dataframe.fillna(1e-10)

        # Get values for groups
        rows = dataframe.index.get_level_values(1).values
        cols = dataframe.columns.get_level_values(1).values

        # Get colors
        row_colors = _get_category_colors(rows, self.colormap_genus)
        col_colors = _get_category_colors(cols, self.colormap_category)

        # Create series with index (for clustermap)
        row_colors = pd.Series(row_colors.values, index=dataframe.index)
        col_colors = pd.Series(col_colors.values, index=dataframe.columns)

        # Plot clustermap
        grid = sns.clustermap(dataframe, center=0.5,
          method=method, metric=metric, vmin=vmin, vmax=vmax, mask=mask,
          cmap=cmap, linewidths=.75, figsize=figsize, xticklabels=True,
          yticklabels=True, row_colors=row_colors, col_colors=col_colors,
          **kwargs)

        # Format labels names
        labelsx = ['(%s) %5s' % (e.get_text().split('-')[1],
                                 e.get_text().split('-')[0])
                        for e in grid.ax_heatmap.get_xticklabels()]
        labelsy = ['%-6s (%s)' % (e.get_text().split('-')[0],
                                 e.get_text().split('-')[1])
                         for e in grid.ax_heatmap.get_yticklabels()]

        # Set labels names
        grid.ax_heatmap.set_xticklabels(labelsx)
        grid.ax_heatmap.set_yticklabels(labelsy, rotation=0)
        #grid.ax_heatmap.set_aspect('equal')

        # Format figure
        plt.suptitle('Antibiogram (%s, %s)' % (metric.lower(),method.lower()))
        plt.subplots_adjust(bottom=0.3, top=0.9, left=0.05, right=0.7)

        # Return
        return grid







if __name__ == '__main__':

    # Configure seaborn style (context=talk)
    sns.set(style="white")

    # Set matplotlib
    mpl.rcParams['xtick.labelsize'] = 9
    mpl.rcParams['ytick.labelsize'] = 9
    mpl.rcParams['axes.titlesize'] = 11
    mpl.rcParams['legend.fontsize'] = 9

    # Pandas configuration
    pd.set_option('display.max_colwidth', 40)
    pd.set_option('display.width', 300)
    pd.set_option('display.precision', 4)

    # Numpy configuration
    np.set_printoptions(precision=2)

    # -------------------------------------------------------------------------
    #                            helper methods
    # -------------------------------------------------------------------------


    # -------------------------------------------------------------------------
    #                                main
    # -------------------------------------------------------------------------
    # Data path.
    path = '../fixtures/fixture_antibiogram.csv'

    # -------------------------------
    # Load data
    # -------------------------------
    # Load
    dataframe = pd.read_csv(path)

    # -------------------------------
    # Create object
    # -------------------------------
    # Antibiogram plotter
    antibiogram = Antibiogram(column_organism='organismCode',
                              column_antibiotic='antibioticCode',
                              column_genus='specieName',
                              column_category='antibioticClass',
                              column_index='sari')

    # Fit antibiogram
    antibiogram = antibiogram.fit(dataframe)

    # ---------
    # Example 1
    # ----------
    antibiogram.plot(organisms=['ECOL', 'SAUR'], figsize=(15, 3))


    # ---------
    # Example 2
    # ---------
    antibiogram.plot(genera=['staphylococcus',
                             'klebsiella',
                             'streptococcus',
                             'enterococcus',
                             'enterobacter'],
                     categories=None,
                     method='weighted',
                     metric='euclidean',
                     figsize=(16, 9))

    # Show
    # plt.show()


##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
#
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division 

# Libraries
import sys
import math
import inspect
import warnings
import itertools
import numpy as np
import pandas as pd

# Libraries.
from scipy.stats import norm
from operator import attrgetter

# Import ARIMA from statsmodels.
from statsmodels.tsa.arima_model import ARIMA

# Libraries wrapper.
from pyamr.core.regression.wreg import RegressionWrapper
from pyamr.core.regression.wreg import _forecast_error
from pyamr.core.regression.wreg import _forecast_conf_int
from pyamr.core.regression.wreg import _insample_conf_int


class ARIMAWrapper(RegressionWrapper):
  """Description


  """

  # --------------------------------------------------------------------------
  #                              overriden
  # --------------------------------------------------------------------------
  def _identifier(self):
    """The name to identify the arima model."""
    return "%s%s [%s, %s]" % (getattr(self, '_name', None),
                              getattr(self, 'order', None),
                              getattr(self, 'trend', None),
                              getattr(self, 'exog', None) is not None)

  # --------------------------------------------------------------------------
  #                         confidence intervals
  # --------------------------------------------------------------------------
  def conf_int_forecast(self, predictions, alpha=0.05):
    """Computes the out-sample confidence interval.

    Parameters
    ----------
    predictions : array-like
      The predictions done by the model.

    alpha : int
      The alpha to configure the confidence interval.

    Returns
    -------
    array-like
    """
    # Compute forecast error
    fcasterr = _forecast_error(sigma2=self._result['sigma2_coef'],
                               arparams=self._raw.arparams, 
                               maparams=self._raw.maparams,
                               steps=predictions.shape[0])
    # Compute forecast interval
    ciou = _forecast_conf_int(forecast=predictions, 
                              fcasterr=fcasterr, 
                              alpha=alpha)
    # Return
    return ciou


  # --------------------------------------------------------------------------
  #                              evaluate
  # --------------------------------------------------------------------------
  def evaluate(self, alpha=0.05):
    """This method creates a dictionary with the relevant attributes.

    .. note:: 
          
          There is an issue with the method conf_int(alpha). It can receive 
          as input parameters a pandas series or a numpy array. We are using 
          it with numpy array as input, hence when pandas series is passed 
          the method _init_result gives an error.

    @see: statsmodels.Arima
    @see: statsmodels.ArimaResults

    Parameters
    ---------- 
    alpha : the confidence interval.

    Returns
    -------
    dictionary : map with all the parameters.
    """
    # Create series.
    d = {} 

    # Add generic metrics.
    d['aic'] = self._raw.aic
    d['bic'] = self._raw.bic
    d['hqic'] = self._raw.hqic
    d['llf'] = self._raw.llf

    # Create params information.
    params_data = zip(self._raw.data.param_names,
                      self._raw.params,
                      self._raw.bse,
                      self._raw.tvalues,
                      self._raw.pvalues,
                      self._raw.conf_int(alpha))

    # Add coefficients statistics to series.
    for name, coef, std, tvalue, pvalue, (cil, ciu) in params_data:
      d['%s_%s'%(name, 'coef')] = coef
      d['%s_%s'%(name, 'std')] = std
      d['%s_%s'%(name, 'tvalue')] = tvalue
      d['%s_%s'%(name, 'tprob')] = pvalue
      d['%s_%s'%(name, 'cil')] = cil
      d['%s_%s'%(name, 'ciu')] = ciu

    # Further statistics
    d.update(self._resid_stats())

    # Extra useful infomation
    d['converged'] = self._raw.mle_retvals['converged']

    # Return
    return d

  # --------------------------------------------------------------------------
  #                             summary method
  # --------------------------------------------------------------------------
  def as_summary(self, **kwargs):
    """This method creates a summary string.
    """ 
    # Variables.
    n = 12 + len(self._raw.params)
    om, omp , dw = 0.0, 0.0, self.m_dw
    jb, jbp = self.m_jb_value, self.m_jb_prob
    nm, nmp = self.m_nm_value, self.m_nm_prob
    skew, kurt = self.m_skew, self.m_kurtosis

    # Summary.
    smry = "\n".join(self._raw.summary(**kwargs).as_text().split("\n")[:n])
    
    # Add in new lines.
    smry+= "\n%s\n%s\n%s\n" % ("="*78, "Manual".center(78, ' '), "-"*78)
    smry+= "Omnibus:       %#22.3f  Durbin-Watson:    %#21.3f\n" % (om, dw)
    smry+= "Prob(Omnibus): %#22.3f  Jarque-Bera (JB): %#21.3f\n" % (omp, jb)
    smry+= "Skew:          %#22.3f  Prob(JB):         %#21.3f\n" % (skew, jbp)
    smry+= "Kurtosis_m:    %#22.3f  Cond No:                 \n" % (kurt) 
    smry+= "Normal (N):    %#22.3f  Prob(N):          %#21.3f\n" % (nm, nmp) 
    smry+= "="*78

    # Return
    return smry
  
  # ---------------------------------------------------------------------------
  #                             prediction method
  # ---------------------------------------------------------------------------
  def get_prediction(self, start=None, end=None, alpha=0.05, **kwargs):
    """This method calls the predict function in ARIMA.

    .. note:: 

      The confidence intervals have been implemented using the functions
      ``_forecast_conf_int()`` and ``_forecast_error()`` implemented in the 
      statsmodels.ts.arima_model.ARIMAResults. As such, it produces the 
      same confidence intervals than the method ``plot_predict()``.

    .. note:: 
    
      The parameter start refers to the original series. As such, if the 
      series has been differenced (e.g. d=1) the first observation has 
      been lost. In those cases, start has to be greater or equal to d.

    .. todo::

      - Review the predictions when dynamic is set to true. 
      - Review the confidence intervals (in and out sample)
      - Review the prediction intervals.

    Parameters
    ----------
    start : int (optional)
      The time t to start the prediction

    end : int (optional)
      The time t to end the prediction

    dynamic : bool
      The dynamic keyword affects in-sample prediction. If dynamic is false
      then the in-sample lagged values are used for prediction. if dynamic
      is true, then in-sample forecasts are used in place of lagged 
      dependent variables. The first forecasted value is start.

    alpha: float
      The alpha value when creating the norm point percent function to
      compute the confidence intervals for the forecast predictions.

    typ: string-like, options = {linear, levels}
      Wether to predict the original levels (levels) or predict in terms of 
      the differenced endogenous variables.

    Returns
    -------
    matrix : 
    """
    # The series has been differenced.
    if hasattr(self._raw.model, 'k_diff'):
      # Return original levels.
      if not 'typ' in kwargs:
        kwargs['typ'] = 'levels' 

    # Compute prediction
    pred = self._raw.predict(start=start, end=end, **kwargs)

    # Compute time.
    time = self._time(start=start, end=end)

    # Divide predictions in insample/forecast
    pred_in = pred[time<len(self.endog)]
    pred_ou = pred[time>=len(self.endog)]

    # Compute confidence intervals
    ciin = self.conf_int_insample(pred_in, alpha=alpha) 
    ciou = self.conf_int_forecast(pred_ou, alpha=alpha)
    cito = np.concatenate((ciin, ciou))

    # Get plotting values.
    return np.column_stack((time, pred, cito)).T


  # ---------------------------------------------------------------------------
  #                               auto method
  # ---------------------------------------------------------------------------
  def auto(self, endog, exog=None, ic='bic', trends=['n','c'], max_ar=3,
                 max_ma=3, max_d=1, warn='ignore', return_fits=False, 
                 converged=True, disp=0, verbose=0, **kwargs):
    """This method finds the best arima through bruteforce.

    Note: It uses grid search through the base wrapper.

    Parameters
    ----------
    endog : array-like
      The endogenous variable (aka. time series data)

    exog : array-like
      The exogenous variable (by default is the time t starting in 0)

    ic : string
      The information criteria which should be any of the params which
      are set in the wrapper (see method _init_result).

    max_ar : number
      The maximum autorregresive order to inspect.

    max_ma : number
      The maximum moving average order to inspect.

    max_d : number
      The maximum difference to inspect.

    warn : string, options = {ignore, raise}
      Whether or not to ignore the warnings raised

    return_fits: bool
      Whether or not to return all the fits. If false only the best model 
      is returned. If true all the models and the best model are returned
      separately.

    converged: bool
      Whether or not to return only converging models. If false all models
      are returned. If true, only those models that converged are returned.

    Returns
    -------
    """
    # Compute all the orders to perform bruteforce search.
    # Note that some orders are not valid, these will throw an
    # exception that will be catched in the base wrapper 
    # grid_search method. For instance (0,0,0) is invalid.
    orders = set(itertools.product(*[np.arange(0,max_ar+1),
                                     np.arange(0,max_d+1),
                                     np.arange(0,max_ma+1)]))

    # Parameters
    grid_params = {'exog': [exog], 
                   'endog': [endog], 
                   'order': orders,
                   'trend': trends, 
                   'disp': [0]}

    # Perform grid search
    with warnings.catch_warnings():
      # What to do with the warnings
      warnings.filterwarnings(warn)
      # Perform gridsearch.
      wrappers = self.grid_search(grid_params=grid_params, verbose=verbose)

    # Keep only those that converged.
    if converged:
      wrappers = [w for w in wrappers if w.converged]

    # Return all
    if return_fits:
      return wrappers, min(wrappers, key=attrgetter(ic))
    
    # Return only best
    return min(wrappers, key=attrgetter(ic))












if __name__ == '__main__': # pragma: no cover

  # Import
  import sys
  import warnings
  import pandas as pd
  import matplotlib as mpl
  import matplotlib.pyplot as plt

  # Filter warnings
  warnings.simplefilter(action='ignore', category=FutureWarning)

  # Set pandas configuration.
  pd.set_option('display.max_colwidth', 14)
  pd.set_option('display.width', 140)
  pd.set_option('display.precision', 4)

  # Import own libraries
  from pyamr.datasets.load import make_timeseries

  # ----------------------------
  # set basic configuration
  # ----------------------------
  # Matplotlib options
  mpl.rc('legend', fontsize=6)
  mpl.rc('xtick', labelsize=6)
  mpl.rc('ytick', labelsize=6)

  # Set pandas configuration.
  pd.set_option('display.max_colwidth', 14)
  pd.set_option('display.width', 150)
  pd.set_option('display.precision', 4)

  # ----------------------------
  # create data
  # ----------------------------
  # Create timeseries data
  x, y, f = make_timeseries()

  # Create exogenous variable
  exog = x

  # ----------------------------
  # fit the model
  # ----------------------------
  # Create specific arima model.
  arima = ARIMAWrapper(estimator=ARIMA).fit(endog=y[:80], 
                                            order=(1,0,0), 
                                            trend='c', 
                                            disp=0)
  
  # Print series
  print("\nSeries:")
  print(arima.as_series())

  # Print summary.
  print("\nSummary:")
  print(arima.as_summary())

  # -----------------
  # Save & Load
  # -----------------
  # File location
  #fname = '../../examples/saved/arima-sample.pickle'

  # Save
  #arima.save(fname=fname)

  # Load
  #arima = ARIMAWrapper().load(fname=fname)


  # -------------
  #  Example I
  # -------------
  # This example shows how to make predictions using the wrapper which has
  # been previously fitted. It also demonstrateds how to plot the resulting
  # data for visualization purposes. It shows two different types of
  # predictions:
  #    - dynamic predictions in which the prediction is done based on the
  #      previously predicted values. Note that for the case of ARIMA(0,1,1)
  #      it returns a line.
  #    - not dynamic in which the prediction is done based on the real
  #      values of the time series, no matter what the prediction was for
  #      those values.

  # Variables.
  s, e = 50, 120

  # Compute predictions
  preds_1 = arima.get_prediction(start=s, end=e, dynamic=False)
  preds_2 = arima.get_prediction(start=s, end=e, dynamic=True)

  # Create figure
  fig, axes = plt.subplots(1, 2, figsize=(8,3))

  # Plot non-dynamic
  # ----------------
  # Plot truth values.
  axes[0].plot(y, color='#A6CEE3', alpha=0.5, marker='o',
                  markeredgecolor='k', markeredgewidth=0.5,
                  markersize=5, linewidth=0.75, label='Observed')

  # Plot forecasted values.
  axes[0].plot(preds_1[0,:], preds_1[1,:], color='#FF0000', alpha=1.00, 
              linewidth=2.0, label=arima._identifier())
  
  # Plot the confidence intervals.
  axes[0].fill_between(preds_1[0,:], preds_1[2,:], 
                                     preds_1[3,:], 
                                     color='#FF0000', 
                                     alpha=0.25)

  # Plot dynamic
  # ------------
  # Plot truth values.
  axes[1].plot(y, color='#A6CEE3', alpha=0.5, marker='o',
                  markeredgecolor='k', markeredgewidth=0.5,
                  markersize=5, linewidth=0.75, label='Observed')

  # Plot forecasted values.
  axes[1].plot(preds_2[0,:], preds_2[1,:], color='#FF0000', alpha=1.00, 
              linewidth=2.0, label=arima._identifier())
  
  # Plot the confidence intervals.
  axes[1].fill_between(preds_2[0,:], preds_2[2,:], 
                                     preds_2[3,:], 
                                     color='#FF0000', 
                                     alpha=0.25)

  # Configure axes
  axes[0].set_title("ARIMA non-dynamic")
  axes[1].set_title("ARIMA dynamic")

  # Format axes
  axes[0].grid(True, linestyle='--', linewidth=0.25)
  axes[1].grid(True, linestyle='--', linewidth=0.25)

  # Legend 
  axes[0].legend()
  axes[1].legend()


  # -------------------------------
  # Example II - AUTO
  # -------------------------------
  # This example shows how to use auto to find the best overall model using
  # a particular seletion criteria. It also demonstrates how to plot the 
  # resulting data for visualization purposes. Note that it only prints
  # the top best classifier according to the information criteria.
  #
  # To do: Review the dynamic=True.
  # Create object
  arima = ARIMAWrapper(estimator=ARIMA)

  # Find the best arima model (bruteforce).
  models, best = arima.auto(endog=y[:80], ic='bic', max_ar=3,  
                                                    max_ma=3, 
                                                    max_d=3,
                                                    return_fits=True)

  # Sort the list (from lower to upper)
  models.sort(key=lambda x: x.bic, reverse=False)

  # Summary
  summary = arima.from_list_dataframe(models)

  # Show summary
  print("\nSummary:")
  print(summary[['arima-order',
                 'arima-trend', 
                 'arima-aic', 
                 'arima-bic']])

  # Create figure
  fig, axes = plt.subplots(3,3, figsize=(10,6))
  axes = axes.flatten()

  # Loop for the selected models
  for i,estimator in enumerate(models[:9]):

    # Show information
    print("%2d. Estimator (bic=%.2f): %s " % \
      (i, estimator.bic, estimator._identifier()))

    # Get the predictions
    preds = estimator.get_prediction(start=s, end=e, dynamic=False)

    # Plot truth values.
    axes[i].plot(y, color='#A6CEE3', alpha=0.5, marker='o',
                    markeredgecolor='k', markeredgewidth=0.5,
                    markersize=5, linewidth=0.75, label='Observed')

    # Plot forecasted values.
    axes[i].plot(preds[0,:], preds[1,:], color='#FF0000', alpha=1.00, 
                 linewidth=2.0, label=estimator._identifier())
    
    # Plot the confidence intervals.
    axes[i].fill_between(preds[0,:], preds[2,:], 
                                     preds[3,:], 
                                     color='#FF0000', 
                                     alpha=0.25)

    # Configure axes
    axes[i].legend(loc=3)
    axes[i].grid(True, linestyle='--', linewidth=0.25)

  # Set superior title
  plt.suptitle("Dynamic predictions for ARIMA")

  # Show
  # plt.show()



###############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
###############################################################################
# Forces decimals on divisions.
from __future__ import division

# Libraries
import sys
import math
import inspect
import numpy as np
import pandas as pd
import statsmodels.api as sm

# External libraries
from scipy.stats import norm
from sklearn.model_selection import ParameterGrid

# Import base wrapper
from pyamr.core.stats.wbase import fargs
from pyamr.core.stats.wbase import BaseWrapper


# ---------------------------------------------------------------------------
#                             helper methods
# ---------------------------------------------------------------------------
def _forecast_error(sigma2, arparams, maparams, steps):
    """This method forecasts the error.

    Used in ARIMAWrapper.

    .. note::

        These functions have been extracted from the ``_forecast_conf_int()``
        and ``_forecast_error()`` implemented in the the class ARIMAResults
        within statsmodels.ts.arima_model. As such, it produces the
        same confidence intervals as those printed using the method
        ``plot_predict()``.

    Parameters
    ----------

    Returns
    -------
    """
    # Libraries.
    from statsmodels.tsa.arima_process import arma2ma
    # Compute error.
    ma_rep = arma2ma(np.r_[1, -arparams],
                     np.r_[1, maparams], lags=steps)  # nobs removed!
    fcasterr = np.sqrt(sigma2 * np.cumsum(ma_rep ** 2))
    # Return
    return fcasterr


def _forecast_conf_int(forecast, fcasterr, alpha=0.05):
    """This method forecasts the confidence interval.

    Parameters
    ----------

    Returns
    -------
    """
    # Libraries.
    from scipy.stats import t, norm
    # Compute confidence intervals
    const = norm.ppf(1 - alpha / 2.)
    conf_int = np.c_[forecast - const * fcasterr,
                     forecast + const * fcasterr]
    # Return
    return conf_int


def _insample_conf_int(forecast, resid, alpha=0.05):
    """This function computes a basic confidence interval.

    Note: It might not be the adecuate way of computing it.

    Parameters
    ----------
    forecast : the forecasted values.
    alpha    : the alpha value selected.

    Returns
    -------
    cilo :
    ciup :
    """
    # Compute variables.
    const = norm.ppf(1.0 - alpha / 2.0)
    mu = np.mean(resid)
    std = np.std(resid)
    c = const * (std / math.sqrt(resid.shape[0]))
    # Compute confidence interval.
    conf_int = np.c_[forecast - c, forecast + c]
    # Return
    return conf_int


class RegressionWrapper(BaseWrapper):
    """Description...


    """
    # Attribute
    _resid = None

    # ---------------------------------------------------------------------------
    #                             HELPER METHODS
    # ---------------------------------------------------------------------------
    def _init_config(self):
        """This method fills self._config with the configuration."""
        # Create dir.
        d = {}
        # Find attributes values for interesting methods.
        d.update(self._getargspecdict(self._raw.model, '__init__'))
        d.update(self._getargspecdict(self._raw.model, 'fit'))
        # Return
        return d

    def _getargspecdict(self, instance, funcname):
        """This method creates a dictionary with pairs name and value.

        Parameters
        ----------
        instance : object with values
        funcname : function which parameters name will be looked for.

        Returns
        -------
        tpls : dictionary with argument name and value.
        """
        try:
            # Get argument parameters.
            func = getattr(instance, funcname, None)
            prms = inspect.getargspec(func)
            tpls = {}
            # Create and fill dictionary
            for name in prms.args:
                if name == 'self': continue
                tpls[name] = getattr(instance, name, None)
            # Return
            return tpls
        except Exception as e:
            # Print
            print("[Exception at _getargspecdict : %s" % e)
            # Return
            return {}

    def conf_int_insample(self, predictions, alpha=0.05):
        """Computes the in-sample confidence interval.

        Parameters
        ----------
        predictions : array-like
          The predictions done by the model.

        alpha : int
          The alpha to configure the confidence interval.

        Returns
        -------
        array-like
        """
        return _insample_conf_int(forecast=predictions,
                                  resid=self._resid,
                                  alpha=alpha)

    # ---------------------------------------------------------------------------
    #              STATISTIC METHODS FOR REGRESSION ANALYSIS
    # ---------------------------------------------------------------------------
    def _resid_stats(self, resid=None, alpha=0.05):
        """This method computes basic stats on the residuals

        Parameters
        ----------
        resid : array-like
          The residuals to perform the stats on.

        alpha : int-like
          The alpha selected.

        Returns
        -------
        dictionary with the stats for the residuals
        """
        # Check if resid is passed.
        if resid is None:
            resid = self._resid

        # No resid to work with.
        if resid is None:
            return {}

        # Create series.
        d = {}

        # Compute autoc-orrelation (durbin-watson)
        from statsmodels.stats.stattools import durbin_watson
        d['m_dw'] = durbin_watson(resid)

        # Compute normalility (jarque bera).
        from statsmodels.stats.stattools import jarque_bera
        jb_value, jb_prob, skew, kurtosis = jarque_bera(resid)
        d['m_jb_value'] = jb_value
        d['m_jb_prob'] = jb_prob
        d['m_skew'] = skew
        d['m_kurtosis'] = kurtosis

        try:
            # Compute normal test (normal test)
            from scipy.stats import normaltest
            nm_value, nm_prob = normaltest(resid)
            d['m_nm_value'] = nm_value
            d['m_nm_prob'] = nm_prob
        except ValueError as e:
            print(e)
            d['m_nm_value'] = np.inf
            d['m_nm_prob'] = np.inf


        # Compute the kolmogorov-smirnov test.
        from scipy.stats import kstest
        ks_value, ks_prob = kstest(resid, 'norm')
        d['m_ks_value'] = ks_value
        d['m_ks_prob'] = ks_prob

        try:
            # Compute the shapiro-wilkinson test.
            from scipy.stats import shapiro
            sh_value, sh_prob = shapiro(resid)
            d['m_shp_value'] = sh_value
            d['m_shp_prob'] = sh_prob
        except ValueError as e:
            print(e)
            d['m_shp_value'] = np.inf
            d['m_shp_prob'] = np.inf

        # Compute anderson-darling.
        # The null hypothesis (sample data is drawn from a population that
        # follows a particular distribution; in this case normal) can be rejected
        # if the statistic es larger than the critical values for an specified
        # significance level.
        from scipy.stats import anderson
        ad_value, ad_cv, ad_sl = anderson(resid)
        d['m_ad_value'] = ad_value
        d['m_ad_nnorm'] = ad_value < ad_cv[2]

        # Return
        return d

    def _exog(self, start=None, end=None):
        """This method generates the exogenous variable time.

        Note: it is only used for those regression methods that do not support
        the parameters start and end in the prediction (wls, ols, rlm, ...).
        On the other side, this method is not necessary for ARIMA since they
        already support this notation.

        .. note:: end is included (see _time()).

        Parameters
        ----------
        start : int (optional)
          The time t to start the prediction

        end : int (optional)
          The time t to end the prediction

        Returns
        -------
        the exogenous variable
        """
        # Default start and end.
        exog = self._time(start=start, end=end)
        # Add constant if required.
        trend = getattr(self, 'trend', None)
        # Return exog without constant
        if trend is None or trend == 'ct':
            return exog
        # Add constant.
        exog = sm.add_constant(exog)
        # Return
        return exog

    def _time(self, start=None, end=None):
        """This method generates the time variable.

        .. note::

          The value indicated by 'end' is included. This is that way
          to match with the implementation of ARIMA from statsmodels.

        Parameters
        ----------
        start : int (optional)
          The time t to start the prediction

        end : int (optional)
          The time t to finish the prediction.

        Returns
        -------
        """
        start = 0 if start is None else start
        end = len(self.endog) if end is None else end + 1
        # Generate time variable.
        return np.arange(start, end, 1)

    def resid(self):
        """Get the residual"""
        if hasattr(self._raw, 'resid'):
            return self._raw.resid

    def fit(self, **kwargs):
        """This method performs the fit.

        Parameters
        ----------
        kwargs : dict-like
          The arguments that will be passed to the method.

        Returns
        -------

        """
        # Empty the class
        self._empty()
        # Update the configuration
        self._config.update(kwargs)

        # Fit the model
        if inspect.isfunction(self.estimator):
            self._fit_funct(**kwargs)
        elif inspect.isclass(self.estimator):
            self._fit_class(**kwargs)

        # Set the residual
        self._resid = self.resid()

        # Evaluate the model
        if self.evaluate:
            self._result = self.evaluate()

        # Return
        return self


if __name__ == '__main__':
    # Constants
    length = 100
    offset = 100
    slope = 10

    # Create time-series.
    x = np.arange(length)
    n = np.random.randn(length) * 10
    y_orig = slope * x + offset + n
    y_pred = slope * x + offset

    # Create and fill a base statistic wrapper.
    w = RegressionWrapper()
    w._resid = y_orig - y_pred

    # Print resid stats
    print(pd.Series(w._resid_stats()))

    # Print resid confidence intervals
    print(w.conf_int_insample(forecast=y_pred))
